"""
Created on: 1/21/2025
Original author: Adil Zaheer
"""

# Built-Ins
import logging
from pathlib import Path

# Third Party
import pandas as pd

# Local Imports
from caf.brain.ml._functions._ml_inputs import (
    DataClassificationInputs,
    ModellingInputs,
    Paths,
    TransformingInputDataInputs,
)
from caf.brain.ml._functions.feature_selection.functions import (
    analyse_feature_importance,
    combine_results,
    rf_feature_selection,
)
from caf.brain.ml._functions.process_data_functions.main import (
    main_input_data,
)

LOG = logging.getLogger(__name__)


def main_feature_selection(
    paths: Paths,
    data_classification: DataClassificationInputs,
    transforming_inputs: TransformingInputDataInputs,
    modelling: ModellingInputs,
    train: pd.DataFrame = None,
    test: pd.DataFrame = None,
    initialised_model=None,
    output: Path | None = None,
):
    """
    Feature selection function.

    Parameters
    ----------
    paths: Path inputs from the PredictionModelInputs class. These inputs
           define paths to external files. See
           caf/brain/ml/main_models/prediction_model/_ml_inputs.py
           for available options.
    data_classification: Data classification inputs from the PredictionModelInputs
                         class. These inputs help define and outline the
                         structure of the input data. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    transforming_inputs: Transforming inputs from the PredictionModelInputs
                         class. These inputs dictate how the data is transformed
                         for machine learning modelling. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    modelling: Modelling inputs from the PredictionModelInputs
               class. These inputs control the machine learning modelling
               pipeline and _functions. See
               caf/brain/ml/main_models/prediction_model/_ml_inputs.py
               for available options.
    train: Transformed input data split into training set.
    test: Transformed input data split into training set.
    initialised_model: Initialised model algorithm from Models enum class.
    output: Path to output location.

    Returns
    -------
    train_final: Dataframe of final training data post feature selection.
    test_final: Dataframe of final test data post feature selection.
    cols_dropped_by_feat_select: These are the columns removed due to
                                 feature selection.
    """

    if modelling.skip_feature_selection:
        return train, test, None

    if output is None:
        if paths.output_path is None:
            raise ValueError("paths.output_path must not be None when output is not provided.")
        output = Path(paths.output_path) / "output"
        output.mkdir(parents=True, exist_ok=True)

    if train is None and test is None:
        LOG.info("Train and test are none so they are being generated by the \
                  main_input_data function")
        data_dict, _, _ = main_input_data(
            paths, data_classification, transforming_inputs, output_path=output
        )

        train = pd.DataFrame.from_dict(data_dict["train_scaled"])
        test = pd.DataFrame.from_dict(data_dict["test_scaled"])

    if modelling.intensive_feature_selection and initialised_model is None:
        LOG.warning("In order to use intensive feature selection, you must \
                     provide an initialised model from the Models class. This \
                     this can be done using the main_model_selection function. \
                     Standard feature selection being ran instead")
        train_final = analyse_feature_importance(
            train_transformed=train,
            target_column=data_classification.target_column,
            weight_column=data_classification.weight_column,
            output_path=output,
            is_time_series=data_classification.is_time_series,
        )

        test_final, cols_dropped_by_feat_select = combine_results(
            train_final=train_final,
            target_column=data_classification.target_column,
            weight_column=data_classification.weight_column,
            test=test,
        )
        return train_final, test_final, cols_dropped_by_feat_select

    if modelling.intensive_feature_selection:
        # eval no. samples
        n_features = (
            len(train.columns)
            - (1 if data_classification.target_column in train.columns else 0)
            - (1 if data_classification.weight_column else 0)
        )
        n_samples = len(train)

        min_samples_per_feature = 15
        required_samples = n_features * min_samples_per_feature

        if n_samples < required_samples:
            LOG.warning("Insufficient data for intensive feature selection. \
                 Falling back to standard feature importance analysis.")
            train_final = analyse_feature_importance(
                train_transformed=train,
                target_column=data_classification.target_column,
                weight_column=data_classification.weight_column,
                output_path=output,
                is_time_series=data_classification.is_time_series,
            )

            test_final, cols_dropped_by_feat_select = combine_results(
                train_final=train_final,
                target_column=data_classification.target_column,
                weight_column=data_classification.weight_column,
                test=test,
            )
            return train_final, test_final, cols_dropped_by_feat_select

        LOG.info("Intensive feature selection running")
        train_final = rf_feature_selection(
            data=train,
            target_column=data_classification.target_column,
            cv=modelling.cv,
            regression_method=initialised_model,
            weight_column=data_classification.weight_column,
            classification_prediction=transforming_inputs.classification_prediction,
            is_time_series=data_classification.is_time_series,
        )

        test_final, cols_dropped_by_feat_select = combine_results(
            train_final=train_final,
            target_column=data_classification.target_column,
            weight_column=data_classification.weight_column,
            test=test,
        )

        return train_final, test_final, cols_dropped_by_feat_select

    LOG.info("Standard feature importance running")
    train_final = analyse_feature_importance(
        train_transformed=train,
        target_column=data_classification.target_column,
        weight_column=data_classification.weight_column,
        output_path=output,
        is_time_series=data_classification.is_time_series,
    )

    test_final, cols_dropped_by_feat_select = combine_results(
        train_final=train_final,
        target_column=data_classification.target_column,
        weight_column=data_classification.weight_column,
        test=test,
    )

    return train_final, test_final, cols_dropped_by_feat_select



"""
Created on: 1/21/2025
Original author: Adil Zaheer
"""

# Built-Ins
import logging
from pathlib import Path

# Third Party
import pandas as pd

# Local Imports
from caf.brain.ml._functions._ml_inputs import (
    DataClassificationInputs,
    ModellingInputs,
    Paths,
    TransformingInputDataInputs,
)
from caf.brain.ml._functions.hparam_optimisation.functions import (
    select_param,
)
from caf.brain.ml._functions.model_selection.main import main_model_selection
from caf.brain.ml._functions.process_data_functions.main import (
    main_input_data,
)

LOG = logging.getLogger(__name__)


def main_hyperparameter_optimisation(
    paths: Paths,
    data_classification: DataClassificationInputs,
    transforming_inputs: TransformingInputDataInputs,
    modelling: ModellingInputs,
    output_folder: Path,
    train: pd.DataFrame = None,
    model_instance=None,
):
    """
    Main function for hyperparameter optimisation.

    Parameters
    ----------
    paths: Path inputs from the PredictionModelInputs class. These inputs
           define paths to external files. See
           caf/brain/ml/main_models/prediction_model/_ml_inputs.py
           for available options.
    data_classification: Data classification inputs from the PredictionModelInputs
                         class. These inputs help define and outline the
                         structure of the input data. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    transforming_inputs: Transforming inputs from the PredictionModelInputs
                         class. These inputs dictate how the data is transformed
                         for machine learning modelling. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    modelling: Modelling inputs from the PredictionModelInputs
               class. These inputs control the machine learning modelling
               pipeline and _functions. See
               caf/brain/ml/main_models/prediction_model/_ml_inputs.py
               for available options.
    train: Dataframe of final training data post feature selection.
    model_instance: Initialised model algorithm from Models enum class.
    output_folder: Path to output location.

    Returns
    -------
    best_model: Fitted final model for prediction on unseen (test) data.
    """
    if train is None:
        LOG.info("Train does not exist so is being generated with the main_input_data \
                  function based on user provided inputs")
        data_dict, _, _ = main_input_data(
            output_path=output_folder,
            paths=paths,
            data_classification=data_classification,
            transforming_inputs=transforming_inputs,
        )

        train = pd.DataFrame.from_dict(data_dict["train_scaled"])

    if model_instance is None:
        LOG.info("Model instance is None so main_model_selection is being called to \
                  obtain the initialised model")
        model_instance = main_model_selection(
            paths=paths,
            data_classification=data_classification,
            transforming_inputs=transforming_inputs,
            modelling=modelling,
            train=train,
            output=output_folder,
        )

    LOG.info("Hyperparameter optimisation underway")
    best_model = select_param(
        train_final=train,
        target_column=data_classification.target_column,
        model_instance=model_instance,
        model_name=modelling.model_choice,
        classification_prediction=transforming_inputs.classification_prediction,
        cv=modelling.cv,
        weight_column=data_classification.weight_column,
        output_folder=output_folder,
        is_time_series=data_classification.is_time_series,
    )
    return best_model



"""
This module contains _functions to conduct data analysis prior to machine
learning model. Data transformation _functions are used if issues are present
and the user permits.
"""

# Built-Ins
import logging
from pathlib import Path
from typing import Optional

# Third Party
import numpy as np
import pandas as pd
import statsmodels.api as sm
from scipy.stats import chi2_contingency, f_oneway, pointbiserialr, shapiro
from sklearn.ensemble import (
    ExtraTreesClassifier,
    GradientBoostingClassifier,
    RandomForestClassifier,
)
from sklearn.linear_model import (
    ElasticNet,
    Lasso,
    LinearRegression,
    LogisticRegression,
    Ridge,
)
from sklearn.multiclass import OneVsRestClassifier
from sklearn.pipeline import Pipeline
from sklearn.tree import DecisionTreeClassifier
from statsmodels.stats.diagnostic import het_breuschpagan, het_white
from statsmodels.stats.outliers_influence import variance_inflation_factor
from statsmodels.stats.stattools import durbin_watson
from statsmodels.tools import add_constant
from statsmodels.tools.sm_exceptions import MissingDataError

# Local Imports
from caf.brain.ml._functions._ml_inputs import DataClassificationInputs, ModellingInputs
from caf.brain.ml._functions.process_data_functions.encode_and_scale import (
    preprocess_numerical_data,
)

LOG = logging.getLogger(__name__)


def pre_forecast_data_analysis(
    output_folder: Path,
    residuals: pd.Series,
    model_fit,
    model_initialised,
    x_test: pd.DataFrame,
    y_test: pd.DataFrame,
    train_scaled: pd.DataFrame,
    test_scaled: pd.DataFrame,
    train_unscaled: pd.DataFrame,
    test_unscaled: pd.DataFrame,
    numerical_pipeline: Optional[Pipeline],
    data_classification: Optional[DataClassificationInputs] = None,
    modelling: Optional[ModellingInputs] = None,
    target_column: Optional[str] = None,
    weight_column: Optional[str] = None,
    numerical_features: Optional[list[str]] = None,
    categorical_features: Optional[list[str]] = None,
    is_time_series: Optional[bool] = None,
    allow_transformations: Optional[bool] = None,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Conduct data analysis for regression or classification problems.

    Parameters
    ----------
    data_classification: Data classification inputs from the PredictionModelInputs
                         class. These inputs help define and outline the
                         structure of the input data. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    modelling: Modelling inputs from the PredictionModelInputs
               class. These inputs control the machine learning modelling
               pipeline and _functions. See
               caf/brain/ml/main_models/prediction_model/_ml_inputs.py
               for available options.
    residuals: Truth values form the train_test_split against the predictions.
    model_fit: Fitted model on train_test_split test data.
    model_initialised: Initialised SciKitLearn model.
    x_test: Dataframe of test data to be used as unseen test data.
    y_test: Dataframe of target data from train, test split.
    train_scaled: Processed input data split into train set.
    test_scaled: Processed input data split into train set.
    train_unscaled: Input data unprocessed split into train.
    test_unscaled: Input data unprocessed split into test.
    numerical_pipeline: Stored numerical transformation pipeline for
                        full model runs. Left as None if not a full
                        model run.
    output_folder: Path to output folder.
    target_column: Name of the target variable column
                   (used if data_classification not provided).
    weight_column: Name of the weight column
                   (used if data_classification not provided).
    numerical_features: List of numerical feature names
                        (used if data_classification not provided).
    categorical_features: List of categorical feature names
                          (used if data_classification not provided).
    is_time_series: Whether data is time series
                    (used if data_classification not provided).
    allow_transformations: Whether to apply transformations
                           (used if modelling not provided).

    Returns
    -------
    Train and test data post data transformations are returned
    if transformations permitted otherwise train and test scaled are returned.
    """

    # extract values from configs if provided
    _target_column = (
        data_classification.target_column if data_classification else target_column
    )
    _weight_column = (
        data_classification.weight_column if data_classification else weight_column
    )
    _numerical_features = (
        data_classification.numerical_features if data_classification else numerical_features
    )
    _categorical_features = (
        data_classification.categorical_features
        if data_classification
        else categorical_features
    )
    _is_time_series = (
        data_classification.is_time_series
        if data_classification
        else (is_time_series or False)
    )
    _allow_transformations = (
        modelling.full_transformations if modelling else (allow_transformations or False)
    )

    is_statsmodel = any(
        base.__module__.startswith("statsmodels")
        for base in model_initialised.__class__.__mro__
    )
    is_classification = isinstance(
        model_fit,
        (
            LogisticRegression,
            GradientBoostingClassifier,
            RandomForestClassifier,
            ExtraTreesClassifier,
            DecisionTreeClassifier,
            OneVsRestClassifier,
        ),
    ) or (
        is_statsmodel
        and ("Logit" in str(model_fit.__class__) or "MNLogit" in str(model_fit.__class__))
    )

    is_linear_model = isinstance(
        model_fit, (LinearRegression, Ridge, Lasso, ElasticNet, LogisticRegression)
    ) or (
        is_statsmodel
        and any(
            name in str(model_fit.__class__) for name in ["OLS", "GLM", "Logit", "MNLogit"]
        )
    )

    if is_classification:
        needs_transformations = _classification_only(
            is_classification=is_classification,
            numerical_features=_numerical_features,
            categorical_features=_categorical_features,
            df=train_unscaled,
            target=_target_column,
            output_path=output_folder,
        )

    else:
        needs_transformations = _regression_only(
            numerical_features=_numerical_features,
            categorical_features=_categorical_features,
            df=train_unscaled,
            target=_target_column,
            output_path=output_folder,
            x_test=x_test,
            y_test=y_test,
            residuals=residuals,
            is_linear_model=is_linear_model,
            is_time_series=is_time_series,
        )

    if needs_transformations:
        if (
            _allow_transformations
            and _numerical_features
            and train_unscaled is not None
            and test_unscaled is not None
        ):
            LOG.info("Transformations applied to numerical data to fix the issues")
            train_out = transform_data(
                df=train_unscaled,
                is_test_data=False,
                numerical_pipeline=numerical_pipeline,
                output_folder=output_folder,
                target_column=_target_column,
                weight_column=_weight_column,
                numerical_features=_numerical_features,
                categorical_features=_categorical_features,
            )
            test_out = transform_data(
                df=test_unscaled,
                is_test_data=True,
                numerical_pipeline=numerical_pipeline,
                output_folder=output_folder,
                target_column=_target_column,
                weight_column=_weight_column,
                numerical_features=_numerical_features,
                categorical_features=_categorical_features,
            )
        else:
            LOG.warning(
                "Data issue present but no numerical features are present or full transformations have \n"
                "not been permitted so transformations can't occur"
            )
            train_out = train_scaled
            test_out = test_scaled
    else:
        LOG.info("No data issue present")
        train_out = train_scaled
        test_out = test_scaled

    return train_out, test_out


def _classification_only(
    is_classification: bool,
    numerical_features: list[str] | None,
    categorical_features: list[str] | None,
    df: pd.DataFrame,
    target: str | None,
    output_path: Path,
) -> bool:
    """
    Classification only problems data analysis.

    Parameters
    ----------
    is_classification:
        True if classification prediction.
    numerical_features:
        List of numerical feature names.
    categorical_features:
        List of categorical feature names.
    df:
        Dataframe containing the data (not encoded or scaled).
    target:
        Name of the target variable column.
    output_path:
        Path to output folder.

    Returns
    -------
    True if issues detected.
    """

    if target is None:
        raise ValueError(
            "Target column must be provided. Ensure target is \n"
            "populated if using data_classification.target in \n"
            "the config file. If calling the pre_forecast_data_analysis\n"
            "function directly, ensure target column is provided."
        )

    numerical_to_fix = []
    categorical_warnings = []
    all_issues = []
    full_path = output_path / "data_analysis_output"
    full_path.mkdir(parents=True, exist_ok=True)

    # NUMERICAL FEATURE CHECKS
    if is_classification and numerical_features:
        if len(set(df[target])) == 2:
            has_issue, weak_feats = _check_numerical_feat_class_target(
                numerical_features=numerical_features,
                df=df,
                target=target,
                output_path=full_path,
            )
            if has_issue:
                numerical_to_fix.extend(weak_feats)
                all_issues.append(
                    {
                        "issue_type": "Weak Numerical-Target Correlation",
                        "severity": "HIGH - Will be transformed",
                        "affected_features": ", ".join(weak_feats),
                        "recommendation": "Log transformation will be applied (if permitted)",
                    }
                )

    if numerical_features:
        has_issue, high_vif_feats = _multicolinearity_check(
            df=df, numerical_features=numerical_features, output_path=full_path
        )
        if has_issue:
            numerical_to_fix.extend(high_vif_feats)
            all_issues.append(
                {
                    "issue_type": "Multicollinearity (VIF > 10)",
                    "severity": "HIGH - Will be transformed",
                    "affected_features": ", ".join(high_vif_feats),
                    "recommendation": "Log transformation will be applied (if permitted)",
                }
            )

    # CATEGORICAL FEATURE CHECKS
    has_imbalance = _class_imbalance_classification_target(
        df=df, target_column=target, output_path=full_path
    )
    if has_imbalance:
        all_issues.append(
            {
                "issue_type": "Class Imbalance",
                "severity": "WARNING - Manual action recommended",
                "affected_features": target,
                "recommendation": "Consider SMOTE, class weights, or stratified sampling",
            }
        )

    if categorical_features:
        has_issue, weak_feats = _check_categorical_classification(
            df=df,
            target=target,
            categorical_features=categorical_features,
            output_path=full_path,
        )
        if has_issue:
            categorical_warnings.extend(weak_feats)
            all_issues.append(
                {
                    "issue_type": "Weak Categorical-Target Relationship (Chi-square p > 0.05)",
                    "severity": "WARNING - Manual action recommended",
                    "affected_features": ", ".join(weak_feats),
                    "recommendation": "Consider feature engineering or removal",
                }
            )

    if categorical_features:
        has_issue, rare_feats = _check_rare_categories(
            df=df, categorical_features=categorical_features, output_path=full_path
        )
        if has_issue:
            categorical_warnings.extend(rare_feats)
            all_issues.append(
                {
                    "issue_type": "Rare Categories (< 5% frequency)",
                    "severity": "WARNING - Manual action recommended",
                    "affected_features": ", ".join(rare_feats),
                    "recommendation": "Consider grouping rare categories or removal",
                }
            )

    if categorical_features:
        has_issue, high_card_feats = _check_cardinality(
            df=df, categorical_features=categorical_features, output_path=full_path
        )
        if has_issue:
            categorical_warnings.extend(high_card_feats)
            all_issues.append(
                {
                    "issue_type": "High Cardinality (> 50 unique values)",
                    "severity": "WARNING - Manual action recommended",
                    "affected_features": ", ".join(high_card_feats),
                    "recommendation": "Consider dimensionality reduction",
                }
            )

    if all_issues:
        pd.DataFrame(all_issues).to_csv(full_path / "data_issues_summary.csv", index=False)
        LOG.warning(
            "Data issues detected. See data_issues_summary.csv in %s for full summary.",
            full_path,
        )
    else:
        LOG.info("No data issues detected.")

    if len(categorical_warnings) > 0:
        categorical_detail = pd.DataFrame(
            {
                "feature": list(set(categorical_warnings)),
                "issue_count": [
                    categorical_warnings.count(feat) for feat in set(categorical_warnings)
                ],
            }
        )
        categorical_detail.to_csv(full_path / "categorical_warnings_detail.csv", index=False)

    return len(all_issues) > 0


def _regression_only(
    df: pd.DataFrame,
    target: str | None,
    numerical_features: list[str] | None,
    categorical_features: list[str] | None,
    output_path: Path,
    x_test: pd.DataFrame = None,
    y_test: pd.DataFrame = None,
    residuals: pd.Series = None,
    is_linear_model: bool = False,
    is_time_series: bool | None = False,
) -> bool:
    """
    Regression only problems data analysis.

    Parameters
    ----------
    target:
        Name of the target variable column.
    output_path:
        Path to output folder.
    numerical_features:
        List of numerical feature names.
    categorical_features:
        List of categorical feature names.
    df:
        Dataframe containing the data (not encoded or scaled).
    x_test:
        Dataframe of test data to be used as unseen test data.
    y_test:
        Dataframe of target data from train, test split.
    residuals:
        Truth values form the train_test_split against the predictions.
    is_linear_model:
        Whether the fitted model is a linear regression model.
    is_time_series:
        If data is temporal in nature.

    Returns
    -------
    True if issues detected.
    """

    if target is None:
        raise ValueError(
            "Target column must be provided. Ensure target is \n"
            "populated if using data_classification.target in \n"
            "the config file. If calling the pre_forecast_data_analysis\n"
            "function directly, ensure target column is provided."
        )

    numerical_to_fix = []
    categorical_warnings = []
    all_issues = []
    full_path = output_path / "data_analysis_output"
    full_path.mkdir(parents=True, exist_ok=True)

    # NUMERICAL FEATURE CHECKS
    if numerical_features:
        has_issue, high_vif_feats = _multicolinearity_check(
            df=df, numerical_features=numerical_features, output_path=full_path
        )
        if has_issue:
            numerical_to_fix.extend(high_vif_feats)
            all_issues.append(
                {
                    "issue_type": "Multicollinearity (VIF > 10)",
                    "severity": "HIGH - Will be transformed",
                    "affected_features": ", ".join(high_vif_feats),
                    "recommendation": "Log transformation will be applied (if permitted)",
                }
            )

    # LINEAR MODEL CHECKS
    if is_linear_model:
        heteroscedasticity_flag = _heteroscedasticity_check(
            x_test=x_test,
            y_test=y_test,
            numerical_features=numerical_features,
            residuals=residuals,
            is_linear_model=is_linear_model,
        )

        linear_model_issues = _linear_model_tests(
            x_test=x_test,
            numerical_features=numerical_features,
            is_time_series=is_time_series,
            residuals=residuals,
            is_linear_model=is_linear_model,
        )

        if heteroscedasticity_flag:
            categorical_warnings.append("Heteroscedasticity")
            all_issues.append(
                {
                    "issue_type": "Heteroscedasticity",
                    "severity": "WARNING - Linear model assumption violated",
                    "affected_features": "All numerical features",
                    "recommendation": "Consider transformation. Applied if permitted",
                }
            )

        if linear_model_issues:
            categorical_warnings.append("Linear model assumptions violated")
            all_issues.append(
                {
                    "issue_type": "Linear Model Assumptions (linearity/normality/autocorrelation)",
                    "severity": "WARNING - Linear model assumption violated",
                    "affected_features": "Model residuals",
                    "recommendation": "Review diagnostic plots; consider non-linear models",
                }
            )

    # CATEGORICAL FEATURE CHECKS
    if categorical_features:
        has_issue, weak_feats = _check_categorical_regression(
            df=df,
            target=target,
            categorical_features=categorical_features,
            output_path=full_path,
        )
        if has_issue:
            categorical_warnings.extend(weak_feats)
            all_issues.append(
                {
                    "issue_type": "Weak Categorical-Target Relationship (ANOVA p > 0.05)",
                    "severity": "WARNING - Manual action recommended",
                    "affected_features": ", ".join(weak_feats),
                    "recommendation": "Consider feature engineering or removal",
                }
            )

        has_issue, rare_feats = _check_rare_categories(
            df=df, categorical_features=categorical_features, output_path=full_path
        )
        if has_issue:
            categorical_warnings.extend(rare_feats)
            all_issues.append(
                {
                    "issue_type": "Rare Categories (< 5% frequency)",
                    "severity": "WARNING - Manual action recommended",
                    "affected_features": ", ".join(rare_feats),
                    "recommendation": "Consider grouping rare categories or removal",
                }
            )

        has_issue, high_card_feats = _check_cardinality(
            df=df, categorical_features=categorical_features, output_path=full_path
        )
        if has_issue:
            categorical_warnings.extend(high_card_feats)
            all_issues.append(
                {
                    "issue_type": "High Cardinality (> 50 unique values)",
                    "severity": "WARNING - Manual action recommended",
                    "affected_features": ", ".join(high_card_feats),
                    "recommendation": "Consider dimensionality reduction",
                }
            )

    if all_issues:
        pd.DataFrame(all_issues).to_csv(full_path / "data_issues_summary.csv", index=False)
        LOG.warning(
            "Data issues detected. See data_issues_summary.csv in %s for full summary.",
            full_path,
        )
    else:
        LOG.info("No data issues detected.")

    if len(categorical_warnings) > 0:
        categorical_detail = pd.DataFrame({"warning": list(set(categorical_warnings))})
        categorical_detail.to_csv(full_path / "categorical_warnings_detail.csv", index=False)

    return len(all_issues) > 0


def transform_data(
    df: pd.DataFrame,
    is_test_data: bool,
    numerical_pipeline,
    output_folder: Path,
    target_column: Optional[str],
    weight_column: Optional[str],
    numerical_features: list[str],
    categorical_features: Optional[list[str]],
) -> pd.DataFrame:
    """
    Function to apply data transformations.

    Parameters
    ----------
    df: Processed input data split into train or test set.
    is_test_data: True if test data being passed.
    numerical_pipeline: Stored numerical transformation pipeline for
                        full model runs. Left as None if not a full
                        model run.
    output_folder: Path to output folder.
    target_column: Name of the target variable column.
    weight_column: Name of the weight column if present.
    numerical_features: List of numerical feature column names.
    categorical_features: List of categorical feature column names.

    Returns
    -------
    transformed_df: Transformed data.
    """

    transformed_data = []

    original_index = df.index

    if target_column and target_column in df.columns:
        target = df[target_column].copy()
    else:
        target = None
    if weight_column and weight_column in df.columns:
        weight = df[weight_column].copy()
    else:
        weight = None

    numerical_data = df[numerical_features].copy()

    # log
    numerical_transformed = np.log1p(numerical_data)

    # force fixing any issues post log transformations
    numerical_transformed = numerical_transformed.replace([np.inf, -np.inf], np.nan)
    numerical_transformed = numerical_transformed.fillna(numerical_transformed.mean())

    # scale
    numerical_scaled, _ = preprocess_numerical_data(
        df=numerical_transformed,
        numerical_features=numerical_features,
        is_test_data=is_test_data,
        numerical_pipeline_train=numerical_pipeline,
        output_folder=output_folder,
    )

    transformed_data.append(numerical_scaled)

    if categorical_features is not None:
        categorical_data = df[categorical_features].copy()
        transformed_data.append(categorical_data)

    if weight is not None:
        transformed_data.append(weight)

    transformed_df = pd.concat(transformed_data, axis=1)

    if target is not None:
        transformed_df[target_column] = target
        transformed_df[target_column] = transformed_df[target_column].astype(int)

    transformed_df.index = original_index

    return transformed_df


def _linear_model_tests(
    x_test: pd.DataFrame,
    numerical_features: list[str] | None,
    is_time_series: bool | None = False,
    residuals: pd.Series = None,
    is_linear_model: bool | None = False,
) -> bool:
    """
    Perform diagnostic checks for linear regression model assumptions.

    This function evaluates three key OLS assumptions:
        - Linearity: whether numerical predictors have a linear relationship
          with the target (checked via correlation with residuals).
        - Normality: whether residuals follow a normal distribution
          (Shapiro–Wilk test).
        - Autocorrelation: whether residuals are independent over time
          (Durbin–Watson test; only applied for time-series data).

    Only applicable to linear regression models with numerical predictors and
    continuous residuals.

    Parameters
    ----------
    x_test:
        Test-set feature matrix.
    numerical_features:
        List of numerical features.
    is_time_series:
        Whether the data has temporal ordering.
    residuals:
        Residuals from the fitted linear model.
    is_linear_model:
        Whether the fitted model is a linear regression model.

    Returns
    -------
    has_violations:
        True if any of the linear model assumptions (linearity, normality,
        autocorrelation) appear to be violated, otherwise False.
    """

    if residuals is None:
        LOG.warning("Residual data is not available, skipping linear model tests.")
        return False

    if not is_linear_model:
        LOG.warning("Model is not linear, skipping linear model tests.")
        return False

    if not numerical_features:
        LOG.info("Numerical features are required, linear model tests, skipping.")
        return False

    if is_linear_model and residuals is not None:
        LOG.info("Running tests for linear model assumptions")
        linearity = False
        normality = False
        autocorrelation = False

        # Linearity
        LOG.info("Checking linearity")
        for col in x_test.columns:
            try:
                correlation = np.corrcoef(x_test[col], residuals)[0, 1]
                if abs(correlation) > 0.1:
                    linearity = True
                    LOG.warning("Warning: %s may not be linearly related to the target.", col)
            except (ValueError, MissingDataError) as e:
                LOG.warning("Linearity test failed for column %s: %s", col, e)

        # Normality
        try:
            if len(residuals) > 5000:
                LOG.info(
                    "Large sample size (%d) - Shapiro-Wilk may be overly sensitive. \n"
                    "Consider results cautiously.",
                    len(residuals),
                )

            _, shapiro_p_value = shapiro(residuals)
            LOG.info("Shapiro-Wilk test p-value: %s", shapiro_p_value)
            if shapiro_p_value < 0.05:
                LOG.warning("Warning: Shapiro-Wilk test suggests non-normality of residuals.")
                normality = True
        except (ValueError, MissingDataError) as e:
            LOG.warning("Shapiro-Wilk test failed: %s", e)

        if is_time_series:
            # Autocorrelation
            try:
                dw_statistic = durbin_watson(residuals)
                LOG.info("Durbin-Watson statistic: %s", dw_statistic)
                if dw_statistic < 1.5 or dw_statistic > 2.5:
                    LOG.warning("Warning: Potential autocorrelation in residuals.")
                    autocorrelation = True
            except (ValueError, MissingDataError) as e:
                LOG.warning("Durbin-Watson test failed: %s", e)

        if autocorrelation or normality or linearity:
            LOG.warning("Linear model specific issues detected.")
            return True

    return False


def _heteroscedasticity_check(
    x_test: pd.DataFrame,
    y_test: pd.DataFrame,
    numerical_features: list[str] | None,
    residuals: pd.Series = None,
    is_linear_model: bool = False,
) -> bool:
    """
    Test for heteroscedasticity using the Breusch–Pagan and White tests.

    Heteroscedasticity refers to non-constant variance of residuals and is a
    violation of the assumptions of ordinary least squares (OLS) regression.
    This diagnostic is only applicable to linear models with numerical predictors.

    Parameters
    ----------
    x_test:
        Test-set feature matrix.
    y_test:
        Test-set target values.
    numerical_features:
        List of numerical features.
    residuals:
        Residuals from the fitted linear model. Required for the tests.
    is_linear_model:
        Whether the fitted model is a linear regression model.

    Returns
    -------
    has_heteroscedasticity:
        True if either the Breusch–Pagan or White test detects heteroscedasticity
        (p < 0.05), otherwise False.
    """

    if residuals is None:
        LOG.warning("Residual data is not available for heteroscedasticity, skipping test.")
        return False

    if not is_linear_model:
        LOG.warning("Model is not linear, skipping heteroscedasticity test.")
        return False

    min_features_for_hetero_test = 3
    if x_test.shape[1] <= min_features_for_hetero_test:
        LOG.info("Skipping heteroscedasticity tests: insufficient features.")
        return False

    if not numerical_features:
        LOG.info(
            "Numerical features are required for heteroscedasticity tests, skipping heteroscedasticity test."
        )
        return False

    LOG.info("Checking for heteroscedasticity")
    x_test_copy = x_test[numerical_features].copy()

    try:
        x_with_const = add_constant(x_test_copy)

        # Check if design matrix is full rank
        matrix_rank = np.linalg.matrix_rank(x_with_const)
        expected_rank = x_with_const.shape[1]

        if matrix_rank < expected_rank:
            LOG.warning(
                "Design matrix is rank deficient (rank %d, expected %d). \n"
                "This is likely due to perfect multicollinearity from encoded categorical variables. \n"
                "Skipping heteroscedasticity tests.",
                matrix_rank,
                expected_rank,
            )
        else:
            try:
                _, bp_test_p_value, _, _ = het_breuschpagan(residuals, x_with_const)
                LOG.info("Breusch-Pagan test p-value: %s", bp_test_p_value)
            except (ValueError, MissingDataError, AssertionError) as e:
                LOG.warning("Breusch-Pagan test failed: %s", e)
                bp_test_p_value = 1.0

            try:
                ols_model = sm.OLS(y_test, x_with_const).fit()
                _, white_test_p_value, _, _ = het_white(ols_model.resid, ols_model.model.exog)
                LOG.info("White's test p-value: %s", white_test_p_value)
            except (
                ValueError,
                MissingDataError,
                AssertionError,
                np.linalg.LinAlgError,
            ) as e:
                LOG.warning("White's test failed: %s", e)
                white_test_p_value = 1.0

            if bp_test_p_value < 0.05 or white_test_p_value < 0.05:
                LOG.warning("Heteroscedasticity detected.")
                return True

    except Exception as e:
        LOG.warning("Heteroscedasticity tests failed: %s", e)
    return False


def _multicolinearity_check(
    df: pd.DataFrame, numerical_features: list[str], output_path: Path
) -> tuple[bool, list[str]]:
    """
    Check numerical features for multicollinearity with Variance Inflation Factor.

    A feature is considered problematic if its VIF exceeds 10, indicating strong
    multicollinearity with other numerical predictors.

    Any problem type, just for numerical x variables.

    Parameters
    ----------
    df:
        Input dataframe containing numerical features.
    numerical_features:
        List of numerical feature names to evaluate.
    output_path:
        Directory where the VIF results CSV will be saved.

    Returns
    -------
    has_multicollinearity:
        True if any features exhibit high VIF (>10), otherwise False.
    high_vif_features:
        List of numerical features with VIF greater than 10.
    """

    high_vif_features: list[str] = []

    if not numerical_features or len(numerical_features) < 2:
        LOG.info("Insufficient numerical features for VIF calculation (need >= 2)")
        return False, high_vif_features

    LOG.info("Checking for multicollinearity")

    x_num = df[numerical_features].copy()
    x_num = x_num.loc[:, x_num.std() > 1e-10]

    if x_num.shape[1] < 2:
        LOG.info("After removing near-constant columns, insufficient features for VIF")
        return False, high_vif_features

    try:
        vif_data = pd.DataFrame()
        vif_data["Feature"] = x_num.columns
        vif_data["VIF"] = [
            variance_inflation_factor(x_num.values, i) for i in range(x_num.shape[1])
        ]

        high_vif = vif_data[vif_data["VIF"] > 10]

        if not high_vif.empty:
            high_vif_features = high_vif["Feature"].tolist()
            LOG.warning("High VIF variables detected: %s", high_vif_features)

        vif_data.to_csv(output_path / "vif_results.csv", index=False)

        return len(high_vif_features) > 0, high_vif_features

    except (ValueError, np.linalg.LinAlgError) as e:
        LOG.warning("VIF calculation failed: %s", e)
        return False, high_vif_features


def _class_imbalance_classification_target(
    df: pd.DataFrame, target_column: str, output_path: Path
) -> bool:
    """
    Check for class imbalance in a classification target.

    Only for classification problems.

    Parameters
    ----------
    df:
        Input dataframe containing the target column.
    target_column:
        Name of the target variable.
    output_path:
        Path to output folder.

    Returns
    -------
    has_imbalance:
        True if class imbalance is detected, otherwise False.
    """

    class_counts = df[target_column].value_counts()
    class_ratios = (class_counts / len(df)).to_dict()
    min_ratio = min(class_ratios.values())
    max_ratio = max(class_ratios.values())

    imbalance_ratio = max_ratio / min_ratio

    has_imbalance = (min_ratio < 0.1) or (imbalance_ratio > 3)
    if has_imbalance:
        LOG.warning("Class imbalance detected. Ratios: %s", class_ratios)
        LOG.warning("Consider: SMOTE, class weights, or stratified sampling")

    class_balance_df = pd.DataFrame(
        {
            "class": class_counts.index,
            "count": class_counts.values,
            "ratio": [class_ratios[c] for c in class_counts.index],
        }
    )
    class_balance_df.to_csv(output_path / "class_balance.csv", index=False)

    return has_imbalance


def _check_cardinality(
    df: pd.DataFrame, categorical_features: list[str], output_path: Path, threshold: int = 50
) -> tuple[bool, list[str]]:
    """
    Check for high cardinality in categorical variables.

    This is where you have too many categories for a categorical variable (e.g.
    1 feature with 2000 categories would cause 2000 dummy columns.).

    Categorical feature check (both problem types).

    Parameters
    ----------
    df:
        Input dataframe.
    categorical_features:
        List of categorical variables to check.
    output_path:
        Path to output file.
    threshold:
        Threshold to check for high cardinality.

    Returns
    -------
    Tuple (bool and the high_card_features)
    """

    high_card_features = []
    cardinality_report = []

    for col in categorical_features:
        n_unique = df[col].nunique()
        cardinality_report.append(
            {
                "feature": col,
                "unique_values": n_unique,
                "cardinality_ratio": n_unique / len(df),
            }
        )

        if n_unique > threshold:
            high_card_features.append(col)
            LOG.warning("High cardinality in %s: %s unique values", col, n_unique)

    pd.DataFrame(cardinality_report).to_csv(
        output_path / "cardinality_report.csv", index=False
    )

    return len(high_card_features) > 0, high_card_features


def _check_rare_categories(
    df: pd.DataFrame,
    categorical_features: list[str],
    output_path: Path,
    threshold: float = 0.05,
) -> tuple[bool, list[str]]:
    """
    Detect rare categories in categorical features.

    A rare category is defined as one whose frequency is below the given
    threshold (default = 5%).

    Categorical feature check (both problem types).

    Parameters
    ----------
    df:
        Input dataframe.
    categorical_features:
        List of categorical variables to check.
    output_path:
        Path to output file.
    threshold:
        Threshold to check for rarity.

    Returns
    -------
    has_rare_categories:
        True if any rare categories were found, otherwise False.
    rare_features:
        List of feature names that contain at least one rare category.
    """
    rare_cats_found = False
    rare_report = []

    for col in categorical_features:
        value_counts = df[col].value_counts(normalize=True)
        rare_values = value_counts[value_counts < threshold]

        if len(rare_values) > 0:
            rare_cats_found = True
            LOG.warning(
                "Rare categories in %s: %s categories <%s percent",
                col,
                len(rare_values),
                (threshold * 100),
            )

            for cat, freq in rare_values.items():
                rare_report.append(
                    {
                        "feature": col,
                        "category": cat,
                        "frequency": freq,
                        "count": (df[col] == cat).sum(),
                    }
                )

    if rare_report:
        pd.DataFrame(rare_report).to_csv(output_path / "rare_categories.csv", index=False)
    rare_features = list({row["feature"] for row in rare_report})
    return rare_cats_found, rare_features


def _check_categorical_classification(
    df: pd.DataFrame, target: str, categorical_features: list[str], output_path: Path
) -> tuple[bool, list[str]]:
    """
    Check the categorical feature versus categorical target relationship.

    Only for classification problem.

    Parameters
    ----------
    df:
        Input dataframe.
    target:
        Target column (y).
    categorical_features:
        List of categorical variables to check.
    output_path:
        Path to output file.

    Returns
    -------
    has_weak_features:
        True if any categorical features show a weak relationship with the
        target.
    weak_features:
        List of categorical feature names that failed the test.
    """
    weak_features = []
    chi_square_results = []

    for col in categorical_features:
        contingency_table = pd.crosstab(df[col], df[target])
        chi2, p_value, _, _ = chi2_contingency(contingency_table)

        chi_square_results.append(
            {
                "feature": col,
                "chi2": chi2,
                "p_value": p_value,
            }
        )

        if p_value > 0.05:
            weak_features.append(col)
            LOG.warning("Weak relationship: %s vs %s (p=%.4f)", col, target, p_value)

    pd.DataFrame(chi_square_results).to_csv(
        output_path / "categorical_chi_square.csv", index=False
    )

    return len(weak_features) > 0, weak_features


def _check_categorical_regression(
    df: pd.DataFrame, target: str, categorical_features: list[str], output_path: Path
) -> tuple[bool, list[str]]:
    """
    One way anova test.

    Only for regression problems and categorical features.

    Parameters
    ----------
    df:
        Input dataframe.
    target:
        Target column (y).
    categorical_features:
        List of categorical variables to check.
    output_path:
        Path to output file.

    Returns
    -------
    has_weak_features:
        True if any categorical features show a weak relationship with the
        target.
    weak_features:
        List of categorical feature names that failed the test.
    """
    weak_features = []
    anova_results = []

    for col in categorical_features:
        groups = [df[df[col] == cat][target].values for cat in df[col].unique()]

        f_stat, p_value = f_oneway(*groups)

        anova_results.append(
            {
                "feature": col,
                "f_statistic": f_stat,
                "p_value": p_value,
            }
        )

        if p_value > 0.05:
            weak_features.append(col)
            LOG.warning("Weak relationship: %s vs %s (p=%.4f)", col, target, p_value)

    pd.DataFrame(anova_results).to_csv(output_path / "categorical_anova.csv", index=False)

    return len(weak_features) > 0, weak_features


def _check_numerical_feat_class_target(
    numerical_features: list[str], df: pd.DataFrame, target: str, output_path: Path
) -> tuple[bool, list[str]]:
    """
    Evaluate relationship between numerical features and a binary classification
    target.

    Numerical features and binary target only.

    Parameters
    ----------
    numerical_features:
        List of numerical features.
    df:
        Input dataframe.
    target:
        Name of the target column.
    output_path:
        Path to output file.

    Returns
    -------
    has_weak_features:
        True if any numerical features show weak correlation with the target.
    weak_features:
        List of numerical feature names identified as weak predictors.
    """
    weak_features = []
    correlations = []

    for col in numerical_features:
        try:
            corr, pval = pointbiserialr(df[target], df[col])
            correlations.append({"feature": col, "correlation": corr, "p_value": pval})
            if abs(corr) < 0.05:
                if pval > 0.05:
                    weak_features.append(col)
        except Exception as e:
            LOG.warning("Feature-target correlation failed for %s: %s", col, e)
            continue

    corr_df = pd.DataFrame(correlations)
    corr_df.to_csv(output_path / "feature_target_correlations.csv", index=False)

    return len(weak_features) > 0, weak_features



"""
Main function to tidy input data.
"""

# Built-Ins
import logging
import pathlib

# Third Party
import pandas as pd

# Local Imports
from caf.brain.ml._functions._ml_inputs import (
    DataClassificationInputs,
    Paths,
    TransformingInputDataInputs,
)
from caf.brain.ml._functions.process_data_functions.encode_and_scale import (
    process_data_pipeline,
)
from caf.brain.ml._functions.process_data_functions.input_data import (
    InitialDataProcessing,
)
from caf.brain.ml._functions.process_data_functions.split_data_into_ttv import (
    split_data,
)

LOG = logging.getLogger(__name__)


def main_input_data(
    paths: Paths,
    data_classification: DataClassificationInputs,
    transforming_inputs: TransformingInputDataInputs,
    output_path,
):
    """
    Main function for processing input data for machine learning modelling.

    Parameters
    ----------
    paths: Path inputs from the PredictionModelInputs class. These inputs
           define paths to external files. See
           caf/brain/ml/main_models/prediction_model/_ml_inputs.py
           for available options.
    data_classification: Data classification inputs from the PredictionModelInputs
                         class. These inputs help define and outline the
                         structure of the input data. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    transforming_inputs: Transforming inputs from the PredictionModelInputs
                         class. These inputs dictate how the data is transformed
                         for machine learning modelling. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    output_path: Path to output location.

    Returns
    -------
    data_dict: Dictionary of processed dataframes.
    drop_vals: Values dropped during encoding of categorical variables.
    numerical_pipeline: SciKit-Learn pipeline that contains scaling information.
    """

    folder_path = pathlib.Path("" if paths.folder_path is None else paths.folder_path)
    output_path = pathlib.Path(output_path)

    if (output_path / "train.csv").exists() or (folder_path / "train.csv").exists():
        try:
            train_raw = pd.read_csv(output_path / "train.csv")
            test_raw = pd.read_csv(output_path / "test.csv")
            try:
                validate = pd.read_csv(output_path / "validate.csv")
                validate[data_classification.target_column] = validate[
                    data_classification.target_column
                ].astype(float)
            except FileNotFoundError:
                LOG.warning("Validate not provided. Validation will not be performed")
                validate = None

        except FileNotFoundError:
            train_raw = pd.read_csv(folder_path / "train.csv")
            test_raw = pd.read_csv(folder_path / "test.csv")
            try:
                validate = pd.read_csv(folder_path / "validate.csv")
                validate[data_classification.target_column] = validate[
                    data_classification.target_column
                ].astype(float)
            except FileNotFoundError:
                LOG.warning("Validate not provided. Validation will not be performed")
                validate = None

        processor = InitialDataProcessing(
            file_path=paths.file_path,
            folder_path=folder_path,
            output_path=output_path,
            target_column=data_classification.target_column,
            custom_index=data_classification.custom_index,
            column_name_to_drop_rows=transforming_inputs.column_name_to_drop_rows,
            value_in_row=transforming_inputs.value_in_row,
            weight_column=data_classification.weight_column,
            categorical_features=data_classification.categorical_features,
            numerical_features=data_classification.numerical_features,
            classification_prediction=transforming_inputs.classification_prediction,
        )

        processed_dfs = {}
        for name, df in [("train", train_raw), ("test", test_raw)]:
            processor.df = df.copy()
            processor.dataframes = {}

            current_name = name
            is_test_data = current_name.lower() == "test"

            processed = processor.data_already_split_pipeline(is_test_data)

            processed_df = list(processed.values())[0]
            processed_dfs[name] = processed_df

            LOG.info("Processed %s dataframe:", name)
            LOG.info("Index names: %s", processed_df.index.names)
            LOG.info("Columns: %s", processed_df.columns.tolist())
            LOG.info("Shape: %s", processed_df.shape)

        train_unscaled = processed_dfs["train"]
        test_unscaled = processed_dfs["test"]
        train_unscaled[data_classification.target_column] = train_unscaled[
            data_classification.target_column
        ].astype(int)

        train_scaled, drop_vals, numerical_pipeline = process_data_pipeline(
            df=train_unscaled.copy(),
            numerical_features=data_classification.numerical_features,
            categorical_features=data_classification.categorical_features,
            target_column=data_classification.target_column,
            weight_column=data_classification.weight_column,
            sample_size_encode=transforming_inputs.sample_size_encode,
            select_encode_values=transforming_inputs.select_encode_values,
            encode_values_to_drop=transforming_inputs.encode_values_to_drop,
            train_encoded=None,
            test_data=False,
            numerical_pipeline=None,
            output_folder=output_path,
        )

        test_scaled, _, _ = process_data_pipeline(
            df=test_unscaled.copy(),
            numerical_features=data_classification.numerical_features,
            categorical_features=data_classification.categorical_features,
            target_column=data_classification.target_column,
            weight_column=data_classification.weight_column,
            sample_size_encode=transforming_inputs.sample_size_encode,
            select_encode_values=transforming_inputs.select_encode_values,
            encode_values_to_drop=transforming_inputs.encode_values_to_drop,
            train_encoded=train_scaled,
            test_data=True,
            numerical_pipeline=numerical_pipeline,
            output_folder=output_path,
        )

        data_dict = {
            "train_scaled": train_scaled,
            "test_scaled": test_scaled,
            "train_unscaled": train_unscaled,
            "test_unscaled": test_unscaled,
            "validate": validate,
        }
        return data_dict, drop_vals, numerical_pipeline

    processor = InitialDataProcessing(
        file_path=paths.file_path,
        folder_path=folder_path,
        output_path=output_path,
        target_column=data_classification.target_column,
        custom_index=data_classification.custom_index,
        column_name_to_drop_rows=transforming_inputs.column_name_to_drop_rows,
        value_in_row=transforming_inputs.value_in_row,
        weight_column=data_classification.weight_column,
        categorical_features=data_classification.categorical_features,
        numerical_features=data_classification.numerical_features,
        classification_prediction=transforming_inputs.classification_prediction,
    )

    is_test_data = False
    processed_dataframes = processor.execute_pipeline(is_test_data)
    if len(processed_dataframes) == 1:
        df = list(processed_dataframes.values())[0]
    else:
        df = pd.concat(processed_dataframes.values(), axis=0)

    train_unscaled, test_unscaled, validate = split_data(
        processed_dataframes=df,
        paths=paths,
        data_classification=data_classification,
        transforming_inputs=transforming_inputs,
        output_path=output_path,
    )

    if validate is not None:
        validate[data_classification.target_column] = validate[
            data_classification.target_column
        ].astype(int)

    train_scaled, drop_vals, numerical_pipeline = process_data_pipeline(
        df=train_unscaled.copy(),
        numerical_features=data_classification.numerical_features,
        categorical_features=data_classification.categorical_features,
        target_column=data_classification.target_column,
        weight_column=data_classification.weight_column,
        sample_size_encode=transforming_inputs.sample_size_encode,
        select_encode_values=transforming_inputs.select_encode_values,
        encode_values_to_drop=transforming_inputs.encode_values_to_drop,
        train_encoded=None,
        test_data=False,
        numerical_pipeline=None,
        output_folder=output_path,
    )

    test_scaled, _, _ = process_data_pipeline(
        df=test_unscaled.copy(),
        numerical_features=data_classification.numerical_features,
        categorical_features=data_classification.categorical_features,
        target_column=data_classification.target_column,
        weight_column=data_classification.weight_column,
        sample_size_encode=transforming_inputs.sample_size_encode,
        select_encode_values=transforming_inputs.select_encode_values,
        encode_values_to_drop=transforming_inputs.encode_values_to_drop,
        train_encoded=train_scaled,
        test_data=True,
        numerical_pipeline=numerical_pipeline,
        output_folder=output_path,
    )

    train_unscaled[data_classification.target_column] = train_unscaled[
        data_classification.target_column
    ].astype(int)
    train_scaled[data_classification.target_column] = train_scaled[
        data_classification.target_column
    ].astype(int)

    data_dict = {
        "train_scaled": train_scaled,
        "test_scaled": test_scaled,
        "train_unscaled": train_unscaled,
        "test_unscaled": test_unscaled,
        "validate": validate,
    }

    return data_dict, drop_vals, numerical_pipeline



"""
Created on: 1/24/2025
Original author: Adil Zaheer
"""

# Built-Ins
from pathlib import Path

# Third Party
import pandas as pd

# Local Imports
from caf.brain.ml._functions.prediction.functions import prediction


def main_prediction(
    model,
    test: pd.DataFrame,
    target_column: str | None,
    output_folder: Path,
    validation: pd.DataFrame,
    weight_column: str | None,
    classification_prediction: tuple[int, ...] | None,
    mse: pd.Series,
    drop_vals: pd.DataFrame,
    cols_dropped_by_feat_select: pd.DataFrame,
) -> None:
    """
    Main prediction function.

    Parameters
    ----------
    model: Fitted final model for prediction on unseen (test) data.
    test: Dataframe of final test data post feature selection.
    target_column: String column name of value to predict.
    output_folder: Path to output location.
    validation: Validation data if available.
    weight_column: Optional string column value to be used as weight.
    classification_prediction: List of integers that correspond to the
                               target column. The value(s) to predict
                               in a classification problem.
    mse: Mean squared error or None. Dependency on if the algorithm selected
         has coefficient values.
    drop_vals: Values dropped during encoding of categorical variables.
    cols_dropped_by_feat_select: These are the columns removed due to
                                 feature selection.

    Returns
    -------
    None
    """

    prediction(
        model=model,
        test=test,
        target_column=target_column,
        output_folder=output_folder,
        validation=validation,
        weight_column=weight_column,
        classification_prediction=classification_prediction,
        mse=mse,
        drop_vals=drop_vals,
        cols_dropped_by_feat_select=cols_dropped_by_feat_select,
    )



"""
Created on: 1/21/2025
Original author: Adil Zaheer
"""

# Built-Ins
import logging
import os
import time
from pathlib import Path
from typing import Optional, Union

# Third Party
import numpy as np
import pandas as pd
import seaborn as sb
from matplotlib import pyplot as plt
from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.feature_selection import RFE, SelectFromModel
from sklearn.inspection import permutation_importance
from sklearn.linear_model import Lasso, LogisticRegression, Ridge
from sklearn.model_selection import (
    BaseCrossValidator,
    KFold,
    RepeatedKFold,
    RepeatedStratifiedKFold,
    StratifiedKFold,
    TimeSeriesSplit,
    cross_val_score,
)
from tqdm import tqdm

# Local Imports
from caf.brain.ml._functions.process_data_functions.split_data_into_ttv import (
    sample_data,
)

LOG = logging.getLogger(__name__)


def rf_feature_selection(
    data: pd.DataFrame,
    target_column: str | None,
    cv: str | None,
    regression_method: BaseEstimator,
    weight_column: str | None,
    classification_prediction: tuple[int, ...] | None,
    is_time_series: bool | None,
) -> pd.DataFrame:
    """
    Two stage feature selection through Random Forest importance and if
    required, a combination of algorithms.

    Parameters
    ----------
    data: Transformed input data split into training set.
    target_column: String column name of value to predict.
    cv: Cross validation method passed as a string. Any popular
        SciKitLearn methods are suitable with KFold being default if
        left as None.
    regression_method: Initialised model algorithm from Models enum class.
    weight_column: Optional string column value to be used as weight.
    classification_prediction: List of integers that correspond to the
                               target column. The value(s) to predict
                               in a classification problem.
    is_time_series: If true then data must be time series. Time series
                    based characteristics are taken into consideration
                    during function execution.


    Returns
    -------
    dataframe_final: Training data post feature selection.

    """
    if not target_column:
        raise ValueError(
            "Please provide a target column for feature selection. \n"
            "This is a column title passed as a string."
        )

    if isinstance(regression_method, LogisticRegression):
        regression_method.set_params(max_iter=1000)
    cv = get_cv_class(cv_method=cv, splits=None, repeats=None, is_time_series=is_time_series)

    x = data.drop(columns=[target_column] + ([weight_column] if weight_column else []))
    y = data[target_column]
    weight = data[weight_column].values.flatten() if weight_column else None
    weight_df = data[weight_column] if weight_column else None

    x_sample, y_sample, weight_sample = sample_data(
        x=x, y=y, weight=weight, is_time_series=is_time_series
    )

    if classification_prediction:
        model = RandomForestClassifier(n_estimators=100, random_state=42, n_jobs=-1)
        n_unique_classes = len(pd.unique(y_sample))
        score_threshold = 0.6
        if n_unique_classes <= 2:
            # binary
            scoring = "accuracy"
        else:
            # multi
            scoring = "f1_weighted"
    else:
        score_threshold = -0.4
        model = RandomForestRegressor(n_estimators=100, random_state=42, n_jobs=-1)
        scoring = "neg_mean_squared_error"

    with tqdm(total=1, desc="Fitting Random Forest") as pbar:
        model.fit(x_sample, y_sample, sample_weight=weight_sample)
        pbar.update(1)

    selector = SelectFromModel(model, prefit=True)
    selected_features = x_sample.columns[selector.get_support()].tolist()

    feature_importance = pd.DataFrame(
        {"feature": x_sample.columns, "importance": model.feature_importances_}
    ).sort_values("importance", ascending=False)

    LOG.info("Feature Importances:")
    for _, row in feature_importance.iterrows():
        LOG.info("%s: %s", row["feature"], row["importance"])

    LOG.info("Number of features selected: %s", len(selected_features))
    if len(selected_features) == 0:
        raise ValueError(
            "No features selected by Random Forest importance threshold. \n"
            "You should evaluate your data as it is likely to be irrelevant for your prediction \n"
            "Running evaluate_data can help diagnose the problem"
        )

    params = {"sample_weight": weight_sample} if weight_sample is not None else {}
    scores = cross_val_score(
        regression_method,
        x_sample[selected_features],
        y_sample,
        cv=cv,
        scoring=scoring,
        n_jobs=-1,
        verbose=0,
        params=params,
    )

    mean_score = np.mean(scores)
    std_score = np.std(scores)

    LOG.info("Cross-validated %s score: %.4f (+/- %.4f)", scoring, mean_score, std_score)

    if classification_prediction:
        needs_intensive = mean_score < score_threshold  # Low accuracy = bad
    else:
        needs_intensive = mean_score > score_threshold

    if needs_intensive:
        LOG.warning(
            "Initial feature selection attempt was inaccurate, trying alternative method"
        )
        result_sample = feature_selection_intensive(
            x=x_sample,
            y=y_sample,
            cv=cv,
            regression_method=regression_method,
            weight=weight_sample,
            classification_prediction=classification_prediction,
        )
        intensive_features = [col for col in result_sample.columns if col != target_column]

        dataframe_final = pd.concat([x[intensive_features], y], axis=1)
        if weight_df is not None:
            dataframe_final = pd.concat([dataframe_final, weight_df], axis=1)
    else:
        dataframe_final = pd.concat([x[selected_features], y], axis=1)
        if weight_df is not None:
            dataframe_final = pd.concat([dataframe_final, weight_df], axis=1)

    return dataframe_final


def feature_selection_intensive(
    x: pd.DataFrame,
    y: pd.Series,
    cv: BaseCrossValidator,
    regression_method,
    weight: Optional[np.ndarray],
    classification_prediction: tuple[int, ...] | None,
) -> pd.DataFrame:
    """
    Thorough feature selection with multiple algorithms.

    Parameters
    ----------
    x: Training data split into explanatory variables only.
    y: Training data split only into the target variable.
    cv: Cross validation method passed as an initialised SciKitLearn CV
        splitter.
    regression_method: Initialised model algorithm from Models enum class.
    weight: Weight values in series form.
    classification_prediction: List of integers that correspond to the
                               target column. The value(s) to predict
                               in a classification problem.

    Returns
    -------
    result: Training data post feature selection.
    """

    if classification_prediction:
        selected_features = _classification_feature_selection(x, y, weight)
        score_threshold = 0.5
    else:
        selected_features = _regression_feature_selection(x, y, weight)
        score_threshold = -1.0  # negative MSE scores

    if len(selected_features) == 0:
        raise ValueError(
            "No features selected by intensive methods. Please \n"
            "Reevaluate your data and what you are trying to predict."
        )

    scores = []
    n_splits = getattr(cv, "n_splits", 5)
    with tqdm(total=n_splits, desc="Cross-validation") as pbar:
        for train_index, test_index in cv.split(x):
            x_train, x_test = x.iloc[train_index], x.iloc[test_index]
            y_train, y_test = y.iloc[train_index], y.iloc[test_index]
            w_train = weight[train_index] if weight is not None else None

            if w_train is not None:
                regression_method.fit(
                    x_train[selected_features], y_train, sample_weight=w_train
                )
            else:
                regression_method.fit(x_train[selected_features], y_train)
            score = regression_method.score(x_test[selected_features], y_test)
            scores.append(score)
            pbar.update(1)
    cv_score = np.mean(scores)
    LOG.info("Number of features selected: %s", len(selected_features))
    LOG.info("Cross-validated score: %s", cv_score)

    if classification_prediction:
        use_all_features = cv_score < score_threshold  # Low score = bad
    else:
        use_all_features = cv_score > score_threshold  # High negative = bad

    if use_all_features:
        LOG.warning("CV score is still not optimal, feature selection is being ignored")
        result = pd.concat([x, y], axis=1)
    else:
        result = pd.concat([x[selected_features], y], axis=1)

    return result


def _classification_feature_selection(
    x: pd.DataFrame, y: pd.Series, weight: np.ndarray | None
) -> list[str]:
    """
    Feature selection algorithms for classification problems.

    Parameters
    ----------
    x: Training data split into explanatory variables only.
    y: Training data split only into the target variable.
    weight: Weight values as numpy array.

    Returns
    -------
    List of selected features based on both algorithms used.
    """
    rfe = RFE(
        estimator=LogisticRegression(random_state=42, max_iter=2000),
        n_features_to_select=min(10, x.shape[1]),
    )
    rfe.fit(x, y)
    rfe_selected = x.columns[rfe.support_].tolist()

    logit_lasso = LogisticRegression(
        penalty="l1", solver="saga", random_state=42, max_iter=2000
    )
    if weight is not None:
        logit_lasso.fit(x, y, sample_weight=weight)
    else:
        logit_lasso.fit(x, y)
    l1_selected = x.columns[abs(logit_lasso.coef_[0]) > 0].tolist()

    final_features = list(set(rfe_selected + l1_selected))
    return final_features


def _regression_feature_selection(
    x: pd.DataFrame, y: pd.Series, weight: np.ndarray | None
) -> list[str]:
    """
    Feature selection algorithms for regression problems.

    Parameters
    ----------
    x: Training data split into explanatory variables only.
    y: Training data split only into the target variable.
    weight: Weight values as numpy array.

    Returns
    -------
    List of selected features based on both algorithms used.
    """
    lasso = Lasso(alpha=0.01, random_state=42)
    if weight is not None:
        lasso.fit(x, y, sample_weight=weight)
    else:
        lasso.fit(x, y)
    lasso_selected = x.columns[abs(lasso.coef_) > 0].tolist()

    ridge = Ridge(alpha=1.0, random_state=42)
    if weight is not None:
        ridge.fit(x, y, sample_weight=weight)
    else:
        ridge.fit(x, y)
    ridge_selected = x.columns[abs(ridge.coef_) > np.mean(abs(ridge.coef_))].tolist()

    final_features = list(set(lasso_selected + ridge_selected))
    return final_features


def get_cv_class(
    cv_method: str | None,
    is_time_series: bool | None,
    splits: int | None = None,
    repeats: int | None = None,
):
    """
    Select which SciKitLearn cross validation method to use.

    Parameters
    ----------
    cv_method: Cross validation method passed as a string. Any popular
               SciKitlearn methods are suitable with KFold being default if
               left as None.
    splits: Number of splits to be used for cross validation.
    repeats: Number of repeats to be used for cross validation.
    is_time_series: If true then data must be time series. Time series
                    based characteristics are taken into consideration
                    during function execution.

    Returns
    -------
    Initialised cross validation method.

    """
    if is_time_series:
        return TimeSeriesSplit(n_splits=splits if splits else 5)

    if cv_method:
        cv_lower = cv_method.lower()
        if cv_lower == "kfold":
            return KFold(n_splits=splits if splits else 5, shuffle=True)
        if cv_lower == "stratifiedkfold":
            return StratifiedKFold(n_splits=splits if splits else 5, shuffle=True)
        if cv_lower == "repeatedkfold":
            return RepeatedKFold(
                n_splits=splits if splits else 5, n_repeats=repeats if repeats else 5
            )
        if cv_lower == "repeatedstratifiedkfold":
            return RepeatedStratifiedKFold(
                n_splits=splits if splits else 5, n_repeats=repeats if repeats else 5
            )
        if cv_lower == "timeseriessplit":
            return TimeSeriesSplit(n_splits=splits if splits else 5)

        LOG.warning("Invalid cross-validation method: %s", cv_method)
        LOG.warning("Using default cross validation method: KFold")
    else:
        LOG.info("No cross-validation method specified, using default: KFold")

    return KFold(n_splits=5, shuffle=True)


def analyse_feature_importance(
    train_transformed: pd.DataFrame,
    target_column: str | None,
    weight_column: str | None,
    output_path: Path | None,
    is_time_series: bool = False,
) -> pd.DataFrame:
    """
    Simple feature selection through importance and correlation metrics with
    results plotted.

    Parameters
    ----------
    train_transformed: Transformed input data split into training set.
    target_column: String column name of value to predict.
    weight_column: Optional string column value to be used as weight.
    output_path: Path to output location.
    is_time_series: If true, data is temporal in nature.

    Returns
    -------
    pd.DataFrame
        - If important features are found: the training data filtered to include
          only selected features plus the target/weight columns.
        - If no features meet importance thresholds: the original training data
          is returned unchanged.
    """
    if not output_path:
        raise ValueError("Please provide an output path")

    pd.set_option("display.float_format", lambda z: f"{z:.10f}")

    if not target_column:
        raise ValueError("A target column is required for feature selection")

    x = train_transformed.drop(
        columns=[target_column] + ([weight_column] if weight_column else [])
    )
    y = train_transformed[target_column]
    weight = train_transformed[weight_column].values.flatten() if weight_column else None

    is_classification = y.dtype == "object" or y.dtype.name == "category" or y.nunique() <= 20
    if is_classification:
        rf = RandomForestClassifier(n_estimators=100, random_state=42, n_jobs=-1)
    else:
        rf = RandomForestRegressor(n_estimators=100, random_state=42, n_jobs=-1)

    x, y, weight = sample_data(x=x, y=y, weight=weight, is_time_series=is_time_series)

    n_rows, _ = train_transformed.shape
    if n_rows > 500000:
        results_df = _analyse_feat_importance_helper(x, y, rf, -1, 3)
    else:
        results_df = _analyse_feat_importance_helper(x, y, rf, -1, 10)

    importance_metrics, filtered_data = filtering_results(
        results_df=results_df,
        target_column=target_column,
        weight_column=weight_column,
        original_data=train_transformed,
    )

    if importance_metrics.empty:
        LOG.warning(
            "All feature importance metrics are zero or near-zero. "
            "This likely indicates insufficient data or data quality issues. "
            "Returning original dataset without feature selection."
        )
        return train_transformed

    if output_path:
        create_importance_plots(results_df=importance_metrics, output_path=output_path)
        results_df.to_csv(
            os.path.join(output_path, "feature_importances.csv"), float_format="%.10f"
        )

    return filtered_data


def _analyse_feat_importance_helper(
    x: pd.DataFrame,
    y: pd.Series,
    rf: Union[RandomForestClassifier, RandomForestRegressor],
    n_jobs: int,
    n_repeats: int,
) -> pd.DataFrame:
    """
    This helper function trains the provided RandomForest model on the given
    feature matrix and target vector, then calculates:
      - Random Forest feature importances
      - Permutation importances (mean and standard deviation)
      - Correlation of each feature with the target

    Parameters
    ----------
    x: Feature matrix used for training and importance calculation.
    y: Target vector corresponding to `x`.
    rf: A scikit-learn RandomForest model instance to fit and evaluate.
    n_jobs: Number of parallel jobs to use for permutation importance.
            (Use -1 to run on all available cores).
    n_repeats: Number of random shuffles to perform for permutation importance.

    Returns
    -------
    DataFrame indexed by feature name containing:
        - importance_rf : RandomForest feature importance scores
        - importance_mean_perm : Mean permutation importance
        - importance_std_perm : Standard deviation of permutation importance
        - correlation : Absolute correlation with the target
    """
    start_time = time.time()

    rf.fit(x, y)
    results = {}

    # Random Forest importance
    importance_df = pd.DataFrame(
        {"feature": x.columns, "importance_rf": rf.feature_importances_}
    ).sort_values("importance_rf", ascending=False)
    results["random_forest_importance"] = importance_df

    # Permutation importance
    perm_importance = permutation_importance(
        rf, x, y, n_repeats=n_repeats, random_state=42, n_jobs=n_jobs
    )

    perm_importance_df = pd.DataFrame(
        {
            "feature": x.columns,
            "importance_mean_perm": perm_importance.importances_mean,
            "importance_std_perm": perm_importance.importances_std,
        }
    ).sort_values("importance_mean_perm", ascending=False)
    results["permutation_importance"] = perm_importance_df

    # Target correlations
    correlations = pd.DataFrame(
        {"feature": x.columns, "correlation": [abs(x[col].corr(y)) for col in x.columns]}
    ).sort_values("correlation", ascending=False)
    results["target_correlations"] = correlations

    results_df = pd.concat(
        [
            importance_df.set_index("feature"),
            perm_importance_df.set_index("feature")[
                ["importance_mean_perm", "importance_std_perm"]
            ],
            correlations.set_index("feature"),
        ],
        axis=1,
    )

    end_time = time.time()
    print("elapsed time:", end_time - start_time)
    return results_df


def filtering_results(
    results_df: pd.DataFrame,
    target_column: str,
    weight_column: str | None,
    original_data: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Helper function for analyse_feature_importance. Results are analysed and
    applied to input data.

    Parameters
    ----------
    results_df: Dataframe of feature selection scores.
    target_column: String column name of value to predict.
    weight_column: Optional string column value to be used as weight.
    original_data: Transformed input data split into training set.

    Returns
    -------
    important_features: Feature selection score results.
    original_data: Training data post feature selection.
    """
    important_features = results_df[
        (results_df["importance_rf"] > 0.05)
        | (results_df["correlation"] > 0.3)
        | (results_df["importance_mean_perm"] > 0.01)
    ]

    selected_features = important_features.index.tolist()
    additional_columns = []
    if target_column:
        additional_columns.append(target_column)
    if weight_column:
        additional_columns.append(weight_column)

    return important_features, original_data[selected_features + additional_columns]


def combine_results(
    train_final: pd.DataFrame,
    target_column: str | None,
    weight_column: str | None,
    test: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Function to apply feature selection results to training data.

    Parameters
    ----------
    train_final: Final training data post feature selection.
    target_column: String column name of value to predict.
    weight_column: Optional string column value to be used as weight.
    test: Transformed input data split into training set.

    Returns
    -------
    test_final: Final test data post feature selection.
    cols_dropped_by_feat_select: Dataframe of explanatory variables
                                 removed during feature selection.
    """
    if target_column in train_final.columns and weight_column in train_final.columns:
        df = train_final.drop(columns=[target_column, weight_column])
    elif target_column in train_final.columns:
        df = train_final.drop(columns=[target_column])
    elif weight_column in train_final.columns:
        df = train_final.drop(columns=[weight_column])
    else:
        df = train_final

    test_final = test[df.columns]

    extra_columns = test.columns.difference(df.columns)
    cols_dropped_by_feat_select = test[extra_columns]

    return test_final, cols_dropped_by_feat_select


def create_importance_plots(results_df: pd.DataFrame, output_path: Path) -> None:
    """
    Plotting feature selection scores where applicable.

    Parameters
    ----------
    results_df: Feature selection score results.
    output_path: Path to output location.

    Returns
    -------
    None
    """
    # rf importance
    plt.figure(figsize=(12, 6))
    sb.barplot(data=results_df.reset_index().head(10), x="importance_rf", y="feature")
    plt.title("Top 10 Features by Random Forest Importance")
    plt.tight_layout()
    plt.savefig(os.path.join(output_path, "rf_importance.png"))
    plt.close()

    # Permutation importance
    plt.figure(figsize=(12, 6))
    results_plot = results_df.reset_index().head(10)
    sb.barplot(
        data=results_plot,
        x="importance_mean_perm",
        y="feature",
    )
    plt.title("Top 10 Features by Permutation Importance")
    plt.tight_layout()
    plt.savefig(os.path.join(output_path, "perm_importance.png"))
    plt.close()

    # correlations
    plt.figure(figsize=(12, 6))
    sb.barplot(data=results_df.reset_index().head(10), x="correlation", y="feature")
    plt.title("Top 10 Features by Correlation with Target")
    plt.tight_layout()
    plt.savefig(os.path.join(output_path, "correlations.png"))
    plt.close()



"""
Created on: 12/16/2024
Original author: Adil Zaheer
"""

# Built-Ins
import enum
from pathlib import Path
from typing import List, Optional, Type, Union

# Third Party
import numpy as np
from caf.toolkit import BaseConfig
from sklearn.base import BaseEstimator
from sklearn.ensemble import (
    AdaBoostRegressor,
    BaggingRegressor,
    ExtraTreesClassifier,
    ExtraTreesRegressor,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import (
    ElasticNet,
    Lasso,
    LinearRegression,
    LogisticRegression,
    Ridge,
)
from sklearn.multiclass import OneVsRestClassifier
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR, LinearSVC
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor


class Models(enum.Enum):
    """
    Available algorithms for the caf.brAIn prediction model.
    """

    LOGIT_REGRESSION_L1 = "logit_regression_l1"
    LOGIT_REGRESSION_L2 = "logit_regression_l2"
    LOGIT_REGRESSION_ELASTICNET = "logit_regression_elasticnet"
    MULTINOMIAL = "multinomial"
    RANDOM_FOREST_REGRESSOR = "random_forest_regressor"
    EXTRA_TREES_REGRESSOR = "extra_trees_regressor"
    GRADIENT_BOOSTING_REGRESSOR = "gradient_boosting_regressor"
    GRADIENT_BOOSTING_CLASSIFIER = "gradient_boosting_classifier"
    ADABOOST_REGRESSOR = "adabost_regressor"
    BAGGING_REGRESSOR = "bagging_regressor"
    SVR = "svr"
    KNN = "knn"
    RIDGE = "ridge"
    LASSO = "lasso"
    ELASTICNET = "elasticnet"
    LINEAR_REGRESSION = "linear_regression"
    DECISION_TREE_REGRESSOR = "decision_tree_regressor"
    RANDOM_FOREST_CLASSIFIER = "random_forest_classifier"
    EXTRA_TREES_CLASSIFIER = "extra_trees_classifier"
    DECISION_TREE_CLASSIFIER = "decision_tree_classifier"
    SVM_CLASSIFIER = "svm_classifier"

    def get_model(self):
        """
        Helper method for algorithm selection and initialisation.

        Returns
        -------
        Tuple of initialised model and empty hyperparameter dictionary.
        """
        model_class, params = MODEL_LOOKUP[self]
        return model_class(**params)


class Paths(BaseConfig):
    """
    Prediction model inputs that are paths to external files.
    """

    file_path: Optional[Path] = None
    folder_path: Optional[Path] = None
    output_path: Optional[Path] = None
    validation_path: Optional[Path] = None


class DataClassificationInputs(BaseConfig):
    """
    Inputs that help define and outline the structure of the input data.
    """

    target_column: Optional[str] = None
    custom_index: Optional[List[str]] = None
    categorical_features: Optional[List[str]] = None
    numerical_features: Optional[List[str]] = None
    weight_column: Optional[str] = None
    is_time_series: Optional[bool] = False


class TransformingInputDataInputs(BaseConfig):
    """
    Inputs that dictate how the data is transformed for machine learning
    modelling.
    """

    column_name_to_drop_rows: Optional[List[str]] = None
    value_in_row: Optional[List[Union[str, int, float]]] = None
    classification_prediction: Optional[tuple[int, ...]] = None
    split_by_value: Optional[str] = None
    split_size: Optional[float] = None
    sample_size_encode: Optional[bool] = False
    select_encode_values: Optional[bool] = False
    encode_values_to_drop: Optional[List[str]] = None


class ModellingInputs(BaseConfig):
    """
    Inputs that control the machine learning modelling pipeline and _functions.
    """

    model_choice: Optional[List[Models]] = None
    full_transformations: Optional[bool] = False
    cv: Optional[str] = None
    skip_feature_selection: Optional[bool] = False
    intensive_feature_selection: Optional[bool] = False


class PredictionModelInputs(BaseConfig):
    """
    Caf.brAIn prediction model inputs.
    """

    paths: Paths = Paths()
    data_classification: DataClassificationInputs = DataClassificationInputs()
    transforming_inputs: TransformingInputDataInputs = TransformingInputDataInputs()
    modelling: ModellingInputs = ModellingInputs()


MODEL_LOOKUP = {
    Models.LOGIT_REGRESSION_L1: (
        LogisticRegression,
        {"penalty": "l1", "solver": "liblinear"},
    ),
    Models.LOGIT_REGRESSION_L2: (
        LogisticRegression,
        {"penalty": "l2"},
    ),
    Models.LOGIT_REGRESSION_ELASTICNET: (
        LogisticRegression,
        {"penalty": "elasticnet", "solver": "saga", "l1_ratio": 0.5},
    ),
    Models.MULTINOMIAL: (
        LogisticRegression,
        {"multi_class": "multinomial", "solver": "lbfgs"},
    ),
    Models.RANDOM_FOREST_REGRESSOR: (
        RandomForestRegressor,
        {},
    ),
    Models.EXTRA_TREES_REGRESSOR: (
        ExtraTreesRegressor,
        {},
    ),
    Models.GRADIENT_BOOSTING_REGRESSOR: (
        GradientBoostingRegressor,
        {},
    ),
    Models.GRADIENT_BOOSTING_CLASSIFIER: (
        GradientBoostingClassifier,
        {},
    ),
    Models.ADABOOST_REGRESSOR: (
        AdaBoostRegressor,
        {},
    ),
    Models.BAGGING_REGRESSOR: (BaggingRegressor, {}),
    Models.SVR: (SVR, {}),
    Models.KNN: (KNeighborsRegressor, {}),
    Models.RIDGE: (Ridge, {}),
    Models.LASSO: (Lasso, {}),
    Models.ELASTICNET: (ElasticNet, {}),
    Models.LINEAR_REGRESSION: (LinearRegression, {}),
    Models.DECISION_TREE_REGRESSOR: (
        DecisionTreeRegressor,
        {},
    ),
    Models.RANDOM_FOREST_CLASSIFIER: (
        RandomForestClassifier,
        {},
    ),
    Models.EXTRA_TREES_CLASSIFIER: (
        ExtraTreesClassifier,
        {},
    ),
    Models.DECISION_TREE_CLASSIFIER: (
        DecisionTreeClassifier,
        {},
    ),
    Models.SVM_CLASSIFIER: (
        OneVsRestClassifier,
        {"estimator": LinearSVC()},
    ),
}


class ModelGrids(enum.Enum):
    """
    Hyperparameter grids for available models from the Models
    enum class.
    """

    RANDOM_FOREST_REGRESSOR = {
        "n_estimators": [10, 50, 100],
        "max_depth": [None, 10, 20, 30],
        "min_samples_split": [2, 5, 10],
        "min_samples_leaf": [1, 2, 4],
    }

    EXTRA_TREES_REGRESSOR = {
        "n_estimators": [50, 100, 200],
        "max_depth": [None, 10, 20, 30],
        "min_samples_split": [2, 5, 10],
        "min_samples_leaf": [1, 2, 4],
    }

    GRADIENT_BOOSTING_REGRESSOR = {
        "n_estimators": [50, 100, 200],
        "learning_rate": [0.01, 0.1, 0.2],
        "max_depth": [3, 4, 5],
    }

    ADABOOST_REGRESSOR = {"n_estimators": [50, 100, 200], "learning_rate": [0.01, 0.1, 0.2]}

    BAGGING_REGRESSOR = {
        "n_estimators": [10, 50, 100],
        "max_samples": [0.5, 0.7, 1.0],
        "max_features": [0.5, 0.7, 1.0],
    }

    SVR = {
        "C": [0.1, 1, 10],
        "kernel": ["linear", "poly", "rbf"],
        "epsilon": [0.1, 0.2, 0.3],
    }

    KNN = {"n_neighbors": [3, 5, 7, 9], "weights": ["uniform", "distance"], "p": [1, 2]}

    RIDGE = {"alpha": [0.1, 1.0, 10.0]}

    LASSO = {"alpha": [0.1, 1.0, 10.0]}

    ELASTICNET = {"alpha": [0.1, 1.0, 10.0], "l1_ratio": [0.1, 0.5, 0.9]}

    DECISION_TREE_REGRESSOR = {
        "max_depth": [None, 10, 20, 30],
        "min_samples_split": [2, 5, 10],
        "min_samples_leaf": [1, 2, 4],
    }

    LOGIT_REGRESSION_L1 = {"C": [1.0, 0.1, 0.01, 0.001]}
    LOGIT_REGRESSION_L2 = {"C": [1.0, 0.1, 0.01, 0.001]}
    LOGIT_REGRESSION_ELASTICNET = {"C": [0.1, 1.0, 10.0], "l1_ratio": [0.1, 0.5, 0.9]}

    MULTINOMIAL = {
        "C": np.arange(0.1, 10, 0.1).tolist(),
        "solver": ["lbfgs", "newton-cg", "sag", "saga"],
    }

    RANDOM_FOREST_CLASSIFIER = {
        "n_estimators": [10, 50, 100, 200],
        "max_depth": [None, 10, 20, 30],
        "min_samples_split": [2, 5, 10],
        "min_samples_leaf": [1, 2, 4],
        "bootstrap": [True, False],
    }

    EXTRA_TREES_CLASSIFIER = {
        "n_estimators": [50, 100, 200],
        "max_depth": [None, 10, 20, 30],
        "min_samples_split": [2, 5, 10],
        "min_samples_leaf": [1, 2, 4],
        "bootstrap": [True, False],
    }

    DECISION_TREE_CLASSIFIER = {
        "max_depth": [None, 10, 20, 30],
        "min_samples_split": [2, 5, 10],
        "min_samples_leaf": [1, 2, 4],
        "criterion": ["gini", "entropy"],
    }

    GRADIENT_BOOSTING_CLASSIFIER = {
        "n_estimators": [50, 100, 200],
        "learning_rate": [0.01, 0.1, 0.3],
        "max_depth": [3, 5, 7],
    }

    SVM_CLASSIFIER = {
        "estimator__C": [0.1, 1, 10],
        "estimator__loss": ["hinge", "squared_hinge"],
    }

    @classmethod
    def get_grid(cls, model_enum):
        return getattr(cls, model_enum.name).value


def get_model_grid_from_type(model_type: Type[BaseEstimator]) -> dict:
    """
    Retrieve the hyperparameter grid for a given scikit-learn model class.

    This function maps a model class (e.g., RandomForestRegressor) to its
    predefined hyperparameter grid stored in the `ModelGrids` enum. It is used
    to support automated hyperparameter tuning workflows.

    Parameters
    ----------
    model_type: The scikit-learn model class for which the hyperparameter grid
                is requested. This should be a class (not an instance).

    Returns
    -------
    A dictionary containing the hyperparameter grid for the specified model class.
    """

    model_type_to_parm_grid = {
        RandomForestRegressor: ModelGrids.RANDOM_FOREST_REGRESSOR.value,
        ExtraTreesRegressor: ModelGrids.EXTRA_TREES_REGRESSOR.value,
        GradientBoostingRegressor: ModelGrids.GRADIENT_BOOSTING_REGRESSOR.value,
        AdaBoostRegressor: ModelGrids.ADABOOST_REGRESSOR.value,
        BaggingRegressor: ModelGrids.BAGGING_REGRESSOR.value,
        SVR: ModelGrids.SVR.value,
        KNeighborsRegressor: ModelGrids.KNN.value,
        Ridge: ModelGrids.RIDGE.value,
        Lasso: ModelGrids.LASSO.value,
        ElasticNet: ModelGrids.ELASTICNET.value,
        DecisionTreeRegressor: ModelGrids.DECISION_TREE_REGRESSOR.value,
        RandomForestClassifier: ModelGrids.RANDOM_FOREST_CLASSIFIER.value,
        ExtraTreesClassifier: ModelGrids.EXTRA_TREES_CLASSIFIER.value,
        DecisionTreeClassifier: ModelGrids.DECISION_TREE_CLASSIFIER.value,
        GradientBoostingClassifier: ModelGrids.GRADIENT_BOOSTING_CLASSIFIER.value,
        LogisticRegression: ModelGrids.LOGIT_REGRESSION_ELASTICNET.value,
    }
    try:
        param_grid = model_type_to_parm_grid[model_type]
        assert isinstance(param_grid, dict)
        return param_grid
    except KeyError as e:
        raise ValueError(f"No grid found for model instance of type {model_type}") from e



"""
Created on: 12/13/2024
Original author: Adil Zaheer
"""

# Built-Ins
import warnings
from abc import ABC, abstractmethod

# Third Party
import pandas as pd


# # # CLASSES # # #
class BaseDataClass(ABC):
    """
    Base class for defining basic DataFrame configuration requirements.
    """

    def __init__(
        self,
        dataframe: pd.DataFrame,
        custom_index: list | None = None,
        target_column: str | None = None,
        column_name_to_drop_rows: str | None = None,
        value_in_row: str | float | int | None = None,
    ):

        self.dataframe = dataframe
        self.custom_index = list() if custom_index is None else custom_index
        self.target_column = target_column
        self.column_name_to_drop_rows = column_name_to_drop_rows
        self.value_in_row = value_in_row

    @abstractmethod
    def index_present(self) -> bool:
        """Validate the presence of an appropriate index"""

    @abstractmethod
    def target_column_present(self) -> bool:
        """Validate the presence of the target column"""

    @abstractmethod
    def explanatory_data(self) -> bool:
        """Validate the presence of explanatory variables"""

    @abstractmethod
    def is_data_numeric(self) -> bool:
        """Validate that the data is numeric"""

    @abstractmethod
    def data_correct_shape(self) -> bool:
        """Validate the shape of the DataFrame"""

    def validate(self) -> bool:
        """Run all validation checks in sequence."""
        self.index_present()
        self.target_column_present()
        self.explanatory_data()
        self.is_data_numeric()
        self.data_correct_shape()
        return True


class ValidateData(BaseDataClass):
    """
    Validation class to ensure data tidying was successful.
    """

    def index_present(self) -> bool:
        """
        Check if a custom index is used or if the default index is acceptable.
        """
        if self.custom_index:

            missing_columns = [
                col for col in self.custom_index if col not in self.dataframe.index.names
            ]
            if missing_columns:
                raise ValueError(f"Custom index columns missing: {missing_columns}")

            if isinstance(self.dataframe.index, pd.MultiIndex):
                index_names = self.dataframe.index.names
                if not all(name in self.custom_index for name in index_names):
                    raise ValueError("MultiIndex does not match custom index")

        return True

    def target_column_present(self) -> bool:
        """
        Verify the presence of the target column.
        """
        if self.target_column is None:
            raise ValueError("Target column not specified")

        if self.target_column not in self.dataframe.columns:
            raise ValueError(f"Target column '{self.target_column}' not present in data")

        return True

    def explanatory_data(self) -> bool:
        """
        Check for explanatory variables.
        """
        explanatory_columns = [
            col for col in self.dataframe.columns if col != self.target_column
        ]

        if len(explanatory_columns) == 0:
            raise ValueError("No explanatory data present")

        if len(explanatory_columns) < 2:
            warnings.warn(
                "Only one explanatory variable present. More variables are generally advised for machine learning modeling."
            )

        return True

    def is_data_numeric(self) -> bool:
        """
        Check if all columns (except target) are numeric.
        """
        non_target_columns = [
            col for col in self.dataframe.columns if col != self.target_column
        ]

        non_numeric_columns = [
            col
            for col in non_target_columns
            if self.dataframe[col].dtype not in ["int64", "float64"]
        ]

        if non_numeric_columns:
            raise ValueError(f"Non-numeric columns found: {non_numeric_columns}")

        return True

    def data_correct_shape(self) -> bool:
        """
        Validate DataFrame structure.
        """
        if self.dataframe.empty:
            raise ValueError("DataFrame is empty")

        if self.target_column not in self.dataframe.columns:
            raise ValueError(f'Target column "{self.target_column}" not found in DataFrame')

        explanatory_columns = [
            col for col in self.dataframe.columns if col != self.target_column
        ]

        if not explanatory_columns:
            raise ValueError("No explanatory variables found in DataFrame")

        return True



"""
Main function used to apply data analysis and corrective measures where
applicable.
"""

# Built-Ins
import logging
import os
from pathlib import Path

# Third Party
import pandas as pd

# Local Imports
from caf.brain.ml._functions._ml_inputs import (
    DataClassificationInputs,
    ModellingInputs,
    Paths,
    TransformingInputDataInputs,
)
from caf.brain.ml._functions.data_analysis.functions import (
    pre_forecast_data_analysis,
)
from caf.brain.ml._functions.model_selection.functions import (
    initialise_model,
)
from caf.brain.ml._functions.process_data_functions.main import (
    main_input_data,
)
from caf.brain.ml._functions.process_data_functions.split_data_into_ttv import (
    simple_train_test_split,
)

LOG = logging.getLogger(__name__)


def main_evaluate_input_data(
    paths: Paths,
    data_classification: DataClassificationInputs,
    transforming_inputs: TransformingInputDataInputs,
    modelling: ModellingInputs,
    output_path: Path | None = None,
    train_scaled: pd.DataFrame = None,
    test_scaled: pd.DataFrame = None,
    train_unscaled: pd.DataFrame = None,
    test_unscaled: pd.DataFrame = None,
    model_fit: object = None,
    model_initialised: object = None,
    residuals: pd.Series = None,
    x_test: pd.DataFrame = None,
    y_test: pd.DataFrame = None,
    numerical_pipeline: object = None,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Main function for testing data quality with traditional statistics

    Parameters
    ----------
    paths: Path inputs from the PredictionModelInputs class. These inputs
           define paths to external files. See
           caf/brain/ml/main_models/prediction_model/_ml_inputs.py
           for available options.
    data_classification: Data classification inputs from the PredictionModelInputs
                         class. These inputs help define and outline the
                         structure of the input data. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    transforming_inputs: Transforming inputs from the PredictionModelInputs
                         class. These inputs dictate how the data is transformed
                         for machine learning modelling. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    modelling: Modelling inputs from the PredictionModelInputs
               class. These inputs control the machine learning modelling
               pipeline and _functions. See
               caf/brain/ml/main_models/prediction_model/_ml_inputs.py
               for available options.
    output_path: Path to output location.
    train_scaled: Processed input data split into train set.
    test_scaled: Processed input data split into train set.
    train_unscaled: Input data unprocessed split into train.
    test_unscaled: Input data unprocessed split into test.
    model_fit: Fitted model on train_test_split test data.
    model_initialised: Initialised SciKitLearn model.
    residuals: Truth values from the train_test_split against the predictions.
    x_test: Dataframe of test data to be used as unseen test data.
    y_test: Dataframe of target data from train, test split.
    numerical_pipeline: Stored numerical transformation pipeline for
                        full model runs. Left as None if not a full
                        model run.
    Returns
    -------
    train_transformed: training data with the numerical features transformed
                       or train_scaled if transformations not applied.
    test_transformed: test data with the numerical features transformed
                      or test_scaled if transformations not applied.
    """
    if not output_path:
        if paths.output_path is None:
            raise ValueError(
                "paths.output_path must not be None when output_path is not provided."
            )
        output_path = paths.output_path / "output"
        if not output_path.exists():
            os.makedirs(output_path)

    if train_scaled is None and train_unscaled is None:
        LOG.info("Evaluation of input data beginning. Outputs are saved to: %s", output_path)

        data_dict, _, numerical_pipeline = main_input_data(
            paths, data_classification, transforming_inputs, output_path=output_path
        )

        train_scaled = pd.DataFrame.from_dict(data_dict["train_scaled"])
        test_scaled = pd.DataFrame.from_dict(data_dict["test_scaled"])
        train_unscaled = pd.DataFrame.from_dict(data_dict["train_unscaled"])
        test_unscaled = pd.DataFrame.from_dict(data_dict["test_unscaled"])
        x_train, x_test, y_train, y_test, x_train_weight = simple_train_test_split(
            df=train_unscaled,
            target_column=data_classification.target_column,
            weight_column=data_classification.weight_column,
        )

        if not modelling.model_choice or len(modelling.model_choice) > 1:
            raise ValueError("Please provide only one algorithm to be used to \
                              evaluate input data if using the function standalone.")
        # model re-initialised as this flow is assuming you aren't running the full model
        model_initialised = modelling.model_choice[0].get_model()

        model_fit, residuals, _ = initialise_model(
            x_train=x_train,
            x_test=x_test,
            y_train=y_train,
            y_test=y_test,
            x_train_weight=x_train_weight,
            output_folder=output_path,
            model_initialised=model_initialised,
            classification_prediction=transforming_inputs.classification_prediction,
        )

        train_transformed, test_transformed = pre_forecast_data_analysis(
            data_classification=data_classification,
            modelling=modelling,
            output_folder=output_path,
            residuals=residuals,
            model_fit=model_fit,
            model_initialised=model_initialised,
            x_test=x_test,
            y_test=y_test,
            train_scaled=train_scaled,
            test_scaled=test_scaled,
            train_unscaled=train_unscaled,
            test_unscaled=test_unscaled,
            numerical_pipeline=numerical_pipeline,
        )
        train_transformed.to_csv(output_path / "train_transformed.csv")
        test_transformed.to_csv(output_path / "test_transformed.csv")

        LOG.info("Evaluation of input data finished.")
        return train_transformed, test_transformed

    train_transformed, test_transformed = pre_forecast_data_analysis(
        data_classification=data_classification,
        modelling=modelling,
        output_folder=output_path,
        residuals=residuals,
        model_fit=model_fit,
        model_initialised=model_initialised,
        x_test=x_test,
        y_test=y_test,
        train_scaled=train_scaled,
        test_scaled=test_scaled,
        train_unscaled=train_unscaled,
        test_unscaled=test_unscaled,
        numerical_pipeline=numerical_pipeline,
    )

    if train_transformed is not None and test_transformed is not None:
        return train_transformed, test_transformed

    return train_scaled, test_scaled



"""
Created on: 1/16/2025
Original author: Adil Zaheer
"""

# Built-Ins
import logging
import time
from pathlib import Path

# Third Party
import joblib
import pandas as pd

# Local Imports
from caf.brain.ml._functions._ml_inputs import PredictionModelInputs
from caf.brain.ml._functions.data_analysis.main import main_evaluate_input_data
from caf.brain.ml._functions.feature_selection.main import main_feature_selection
from caf.brain.ml._functions.hparam_optimisation.main import (
    main_hyperparameter_optimisation,
)
from caf.brain.ml._functions.model_selection.functions import initialise_model
from caf.brain.ml._functions.model_selection.main import main_model_selection
from caf.brain.ml._functions.prediction.main import main_prediction
from caf.brain.ml._functions.process_data_functions.main import main_input_data
from caf.brain.ml._functions.process_data_functions.split_data_into_ttv import (
    simple_train_test_split,
)

LOG = logging.getLogger(__name__)


def main(params: PredictionModelInputs, output_path: Path) -> None:
    """
    The main function for the caf.brAIn full machine learning pipeline.

    The prediction model utilises machine learning libraries in order to
    generate predictions. It's designed to streamline the process and remove
    any barrier to entry thereby making machine learning modelling more
    accessible.

    Parameters
    ----------
    params: config file inputs
    output_path: path to output file location. Should be generated during
                 model setup if not passed directly.

    """
    start_time = time.time()

    paths = {
        "train_scaled": Path(output_path) / "train_scaled.csv",
        "test_scaled": Path(output_path) / "test_scaled.csv",
        "train_unscaled": Path(output_path) / "train_unscaled.csv",
        "test_unscaled": Path(output_path) / "test_unscaled.csv",
        "validate": Path(output_path) / "validate.csv",
    }

    if paths["train_scaled"].exists() and paths["test_scaled"].exists():
        LOG.info(
            "Training data is already present from a previous model run \n"
            "so it is being read in."
        )

        drop_vals_path = Path(output_path) / "dropped_encoding_vals.csv"
        if drop_vals_path.exists():
            drop_vals = pd.read_csv(drop_vals_path)
        else:
            drop_vals = None

        data = {
            name: pd.read_csv(path) if path.exists() else None for name, path in paths.items()
        }

        validate = data["validate"]

        custom_index = params.data_classification.custom_index
        if custom_index:
            for key in ["train_scaled", "test_scaled", "train_unscaled", "test_unscaled"]:
                df = data[key]
                if df is not None and all(col in df.columns for col in custom_index):
                    data[key] = df.set_index(custom_index, verify_integrity=False)

            if validate is not None and all(col in validate.columns for col in custom_index):
                validate = validate.set_index(custom_index, verify_integrity=False)

        train_scaled = data["train_scaled"]
        test_scaled = data["test_scaled"]
        train_unscaled = data["train_unscaled"]
        test_unscaled = data["test_unscaled"]

        if params.data_classification.numerical_features:
            num_pipe_path = Path(output_path) / "numerical_pipeline.pkl"
            numerical_pipeline = joblib.load(num_pipe_path)
        else:
            numerical_pipeline = None

    else:
        LOG.info(
            "Training data is not present from a previous model run so it \n"
            "is being generated."
        )

        data_dict, drop_vals, numerical_pipeline = main_input_data(
            output_path=output_path,
            paths=params.paths,
            data_classification=params.data_classification,
            transforming_inputs=params.transforming_inputs,
        )

        train_scaled = pd.DataFrame.from_dict(data_dict["train_scaled"])
        test_scaled = pd.DataFrame.from_dict(data_dict["test_scaled"])
        train_unscaled = pd.DataFrame.from_dict(data_dict["train_unscaled"])
        test_unscaled = pd.DataFrame.from_dict(data_dict["test_unscaled"])

        train_scaled.to_csv(paths["train_scaled"], index=True)
        test_scaled.to_csv(paths["test_scaled"], index=True)
        train_unscaled.to_csv(paths["train_unscaled"], index=True)
        test_unscaled.to_csv(paths["test_unscaled"], index=True)

        validate = None
        if data_dict["validate"] is not None and len(data_dict["validate"]) > 0:
            validate = pd.DataFrame.from_dict(data_dict["validate"])
            validate.to_csv(paths["validate"], index=True)

    selected_model = main_model_selection(
        paths=params.paths,
        data_classification=params.data_classification,
        transforming_inputs=params.transforming_inputs,
        modelling=params.modelling,
        train=train_scaled,
        output=output_path,
    )

    x_train, x_test, y_train, y_test, x_train_weight = simple_train_test_split(
        df=train_scaled,
        target_column=params.data_classification.target_column,
        weight_column=params.data_classification.weight_column,
    )

    x_train_model_fit, residuals, mse = initialise_model(
        x_train=x_train,
        x_test=x_test,
        y_train=y_train,
        y_test=y_test,
        x_train_weight=x_train_weight,
        output_folder=output_path,
        model_initialised=selected_model,
        classification_prediction=params.transforming_inputs.classification_prediction,
    )

    train_transformed, test_transformed = main_evaluate_input_data(
        paths=params.paths,
        data_classification=params.data_classification,
        transforming_inputs=params.transforming_inputs,
        modelling=params.modelling,
        output_path=output_path,
        train_scaled=train_scaled,
        test_scaled=test_scaled,
        train_unscaled=train_unscaled,
        test_unscaled=test_unscaled,
        model_fit=x_train_model_fit,
        model_initialised=selected_model,
        residuals=residuals,
        x_test=x_test,
        y_test=y_test,
        numerical_pipeline=numerical_pipeline,
    )

    train_final, test_final, cols_dropped_by_feat_select = main_feature_selection(
        paths=params.paths,
        data_classification=params.data_classification,
        transforming_inputs=params.transforming_inputs,
        modelling=params.modelling,
        train=train_transformed,
        test=test_transformed,
        output=output_path,
        initialised_model=selected_model,
    )

    best_model = main_hyperparameter_optimisation(
        paths=params.paths,
        data_classification=params.data_classification,
        transforming_inputs=params.transforming_inputs,
        modelling=params.modelling,
        train=train_final,
        model_instance=selected_model,
        output_folder=output_path,
    )

    main_prediction(
        model=best_model,
        test=test_final,
        target_column=params.data_classification.target_column,
        output_folder=output_path,
        validation=validate,
        weight_column=params.data_classification.weight_column,
        classification_prediction=params.transforming_inputs.classification_prediction,
        mse=mse,
        drop_vals=drop_vals,
        cols_dropped_by_feat_select=cols_dropped_by_feat_select,
    )

    end_time = time.time()
    LOG.info("Total run time: %.2f seconds", (end_time - start_time))



"""
caf.brAIn is Transport for the North's bespoke machine learning and AI
library. It consists of a generalised end to end machine learning pipeline,
simplified machine learning functions and a machine vision model.
"""

from caf.brain.ml._ml import (
    tidy_data,
    transform_data,
    feature_selection,
    algorithm_evaluation,
    hparam_optim,
    evaluate_data,
)

from caf.brain.ml._functions._ml_inputs import Models



"""Created on: 9/9/2025. Original author: Adil Zaheer"""

# Built-Ins
import logging
import os.path
from pathlib import Path
from typing import Optional

# Third Party
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.linear_model import ElasticNet, LogisticRegression

# Local Imports
from caf.brain.ml._functions._baseclasses import ValidateData
from caf.brain.ml._functions._ml_inputs import Models
from caf.brain.ml._functions.data_analysis.functions import (
    pre_forecast_data_analysis,
)
from caf.brain.ml._functions.feature_selection.functions import (
    analyse_feature_importance,
)
from caf.brain.ml._functions.hparam_optimisation.functions import (
    select_param,
)
from caf.brain.ml._functions.model_selection.functions import (
    initialise_model,
    select_model,
)
from caf.brain.ml._functions.process_data_functions.encode_and_scale import (
    process_data_pipeline,
)
from caf.brain.ml._functions.process_data_functions.input_data import (
    InitialDataProcessing,
)
from caf.brain.ml._functions.process_data_functions.split_data_into_ttv import (
    simple_train_test_split,
    stratified_split_with_categories,
)

LOG = logging.getLogger(__name__)


def _load_data(data: Optional[pd.DataFrame], data_path: Path | None) -> pd.DataFrame:
    """
    Loads data into a pandas dataframe.

    Parameters
    ----------
    data
        Pandas Dataframe of your data. Structured or semi-structured tabular
        format.
    data_path
        Path to your structured or semi-structured tabular data.

    Returns
    -------
    Pandas dataframe.
    """
    if data is not None:
        return data
    if data_path:
        return pd.read_csv(data_path)
    raise ValueError("No data or data_path provided")


def tidy_data(
    data_path: Path | None,
    classification_prediction: tuple[int, ...] | None,
    output_path: Path | str,
    categorical_features: list[str] | None,
    numerical_features: list[str] | None,
    target: str,
    custom_index: list[str] | None = None,
    weight: str | None = None,
    column_name_to_drop_rows: Optional[list[str]] = None,
    value_in_row: list[str | float | int] | None = None,
    data: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """
    Converts semi-structured data into structured for machine learning.

    Parameters
    ----------
    data_path
        Path to your structured or semi-structured tabular data.
    classification_prediction
        List of integers that correspond to the target column. The value(s) to
        predict in a classification problem.
    output_path
        Path to output location.
    categorical_features
        List of column names (strings) that are categorical variables.
    numerical_features
        List of column names (strings) that are continuous variables.
    custom_index
        Columns in your data that are to be indexed e.g. year, geography.
    target
        Column in your data that is the target variable (Y, dependent variable),
        what you want to predict.
    weight
        Optional string column value to be used as weight.
    column_name_to_drop_rows
        List of string column names that contain values to drop.
    value_in_row
        Corresponding values for column_name_to_drop_rows.
    data
        Pandas Dataframe of your data. Structured or semi-structured tabular
        format.

    Returns
    -------
    Structured dataframe.
    """
    output_path = Path(output_path)

    dataframe = _load_data(data=data, data_path=data_path)
    processor = InitialDataProcessing(
        file_path=None,
        folder_path=None,
        output_path=output_path,
        target_column=target,
        custom_index=custom_index,
        column_name_to_drop_rows=column_name_to_drop_rows,
        value_in_row=value_in_row,
        weight_column=weight,
        categorical_features=categorical_features,
        numerical_features=numerical_features,
        classification_prediction=classification_prediction,
    )
    processor.df = dataframe
    processed = processor.data_already_split_pipeline(is_test_data=False)
    processed_df = list(processed.values())[0]
    if output_path:
        processed_df.to_csv(os.path.join(output_path, "tidy_data.csv"))
        LOG.info("Tidy data output as output path provided")
    return processed_df


def transform_data(
    data: pd.DataFrame,
    output_path: Path | str,
    categorical_features: list[str] | None,
    numerical_features: list[str] | None,
    target: str,
    custom_index: list[str] | None = None,
    weight: str | None = None,
    sample_size_encode: bool | None = None,
    select_encode_values: bool | None = None,
    encode_values_to_drop: list[str] | None = None,
) -> pd.DataFrame:
    """
    Encode and or scale data where applicable for machine learning modelling.

    Parameters
    ----------
    data
        Pandas Dataframe of your data. Structured or semi-structured tabular
        format.
    output_path
        Path to output location.
    categorical_features
        List of column names (strings) that are categorical variables.
    numerical_features
        List of column names (strings) that are continuous variables.
    target
        Column in your data that is the target variable (Y, dependent variable),
        what you want to predict.
    custom_index
        Columns in your data that are to be indexed e.g. year, geography.
    weight
        Optional string column value to be used as weight.
    sample_size_encode
        Optional bool. If true, the data will be split based on sample size.
        Variables with the largest sample size will be used as reference class.
    select_encode_values
        Optional bool. If True, data is split based on custom values set by the
        user. Corresponds to encode_values_to_drop.
    encode_values_to_drop
        If select_encode_values is True, then this must be a list of strings
        the length of categorical_features. Position one in the list will link
        to the first variable provided in categorical_features and so on.

    Returns
    -------
    Pandas dataframe of input data encoded and scaled.

    """
    output_path = Path(output_path)

    if not ValidateData(
        dataframe=data, custom_index=custom_index, target_column=target
    ).validate():
        raise ValueError(
            "Data not suitable for encoding and scaling. Please run \n"
            "full model flow or tidy_data method prior to \n"
            "data analysis."
        )
    if not output_path:
        raise ValueError("Please provide an output path to use the _transform_data")

    preprocessed_df, _, _ = process_data_pipeline(
        df=data,
        numerical_features=numerical_features,
        categorical_features=categorical_features,
        target_column=target,
        weight_column=weight,
        sample_size_encode=sample_size_encode,
        select_encode_values=select_encode_values,
        encode_values_to_drop=encode_values_to_drop,
        train_encoded=None,
        test_data=False,
        numerical_pipeline=None,
        output_folder=output_path,
    )

    preprocessed_df.to_csv(os.path.join(output_path, "transformed_data.csv"))
    LOG.info("Transformed data output %s", output_path)
    return preprocessed_df


def feature_selection(
    data: pd.DataFrame,
    output_path: Path | str,
    categorical_features: list[str] | None,
    numerical_features: list[str] | None,
    target: str,
    custom_index: list[str] | None = None,
    is_encoded: bool = True,
    weight: str | None = None,
    sample_size_encode: bool | None = None,
    select_encode_values: bool | None = None,
    encode_values_to_drop: list[str] | None = None,
) -> pd.DataFrame:
    """
    Conduct simple feature selection.

    Parameters
    ----------
    is_encoded
        is data encoded and scaled.
    data
        Pandas Dataframe of your data. Structured or semi-structured tabular
        format.
    output_path
        Path to output location.
    target
        Column in your data that is the target variable (Y, dependent variable),
        what you want to predict.
    weight
        Optional string column value to be used as weight.
    custom_index
        Columns in your data that are to be indexed e.g. year, geography.
    categorical_features
        List of column names (strings) that are categorical variables.
    numerical_features
        List of column names (strings) that are continuous variables.
    sample_size_encode
        Optional bool. If true, the data will be split based on sample size.
        Variables with the largest sample size will be used as reference class.
    select_encode_values
        Optional bool. If True, data is split based on custom values set by
        the user. Corresponds to encode_values_to_drop.
    encode_values_to_drop
        If select_encode_values is True, then this must be a list of strings
        the length of categorical_features. Position one in the list will link
        to the first variable provided in categorical_features and so on.

    Returns
    -------
    df_final
        Feature selected dataset.
    """
    output_path = Path(output_path)

    if not ValidateData(
        dataframe=data, custom_index=custom_index, target_column=target
    ).validate():
        raise ValueError(
            "Data not suitable for data analysis. Please run \n"
            "full model flow or tidy_data method prior to \n"
            "data analysis."
        )

    LOG.warning(
        "Data should already been encoded and scaled where applicable \n"
        "If this is not the case set is_encoded to false"
    )

    if not is_encoded:
        df = transform_data(
            data=data,
            output_path=output_path,
            categorical_features=categorical_features,
            numerical_features=numerical_features,
            target=target,
            custom_index=custom_index,
            sample_size_encode=sample_size_encode,
            select_encode_values=select_encode_values,
            encode_values_to_drop=encode_values_to_drop,
        )
    else:
        df = data

    df_final = analyse_feature_importance(
        train_transformed=df,
        target_column=target,
        weight_column=weight,
        output_path=output_path,
    )
    return df_final


def algorithm_evaluation(
    model_choice: list[Models] | Models,
    data: pd.DataFrame,
    output_path: Path | str,
    target: str,
    custom_index: list[str] | None = None,
    weight: str | None = None,
    classification_prediction: tuple[int, ...] | None = None,
) -> BaseEstimator:
    """
    Evaluate which algorithm is best performing.

    Algorithms must be from the Models enum class. It is advised to ensure data
    is in an optimal state in order to get accurate results. This means data is
    encoded and scaled where applicable and feature selection is applied. This
    can be done with:

    ``from caf.brain.ml import feat_selection, transform_data, algorithm_evaluation``

    Parameters
    ----------
    model_choice
        List or one algorithm to use as the base of the model.
        Available algorithms can be seen in _ml_inputs.py or __info__.py.
    data
        Pandas Dataframe of your data. Structured or semi-structured
        tabular format.
    output_path
        Path to output location.
    target
        Column in your data that is the target variable (Y, dependent
        variable), what you want to predict.
    weight
        Optional string column value to be used as weight.
    classification_prediction
        List of integers that correspond to the target column. The value(s) to
        predict in a classification problem.
    custom_index
        Columns in your data that are to be indexed e.g. year, geography etc.
    Returns
    -------
    Initialised best performing model.
    """
    output_path = Path(output_path)

    LOG.warning(
        "It is advised to run both transform_data and feat_selection \n"
        "prior to algorithm_evaluation."
    )

    if not ValidateData(
        dataframe=data, custom_index=custom_index, target_column=target
    ).validate():
        raise ValueError(
            "Data not suitable for algorithm evaluation. Please run \n"
            "full model flow or tidy_data method prior to \n"
            "algorithm evaluation."
        )
    if not output_path:
        raise ValueError("Please provide an output path to use algorithm_evaluation")

    if not isinstance(model_choice, list):
        model = [model_choice]
    else:
        model = model_choice

    selected_model = select_model(
        train=data,
        target_column=target,
        weight_column=weight,
        models_to_test=model,
        classification_prediction=classification_prediction,
        output_folder=output_path,
    )

    return selected_model


def hparam_optim(
    model_choice: Models,
    data: pd.DataFrame,
    output_path: Path | str,
    target: str,
    is_time_series: bool | None = None,
    weight: str | None = None,
    classification_prediction: tuple[int, ...] | None = None,
) -> BaseEstimator:
    """
    Hyperparameter optimisation for your selected algorithm.

    Algorithm must be part of the Models enum class.

    Parameters
    ----------
    model_choice
        List or one algorithm to use as the base of the model.
        Available algorithms can be seen in _ml_inputs.py or __info__.py.
    is_time_series
        If true then data must be time series. Time series based
        characteristics are taken into consideration during function execution.
    data
        Pandas Dataframe of your data. Structured or semi-structured tabular
        format.
    output_path
        Path to output location.
    target
        Column in your data that is the target variable (Y, dependent variable),
        what you want to predict.
    weight
        Optional string column value to be used as weight.
    classification_prediction
        List of integers that correspond to the target column. The value(s) to
        predict in a classification problem.
    Returns
    -------
    Initialised model with the best combination of hyperparameters.

    """
    if not output_path:
        raise ValueError("Please provide an output path to use hparam_optim")

    output_path = Path(output_path)

    if not isinstance(model_choice, list):
        model = [model_choice]
    else:
        model = model_choice

    if len(model) > 1:
        LOG.warning(
            "More than one model selected. The first model will be \n"
            "optimised. To find the best performing model, call algorithm_evaluation or \n"
            "main_model_selection"
        )

    LOG.warning(
        "Data should be encoded and scaled where applicable. Call \n"
        "_transform_data to do this prior to hyperparameter optimisation"
    )

    selected_model = model[0].get_model()

    final_model = select_param(
        train_final=data,
        target_column=target,
        model_instance=selected_model,
        model_name=model_choice,
        classification_prediction=classification_prediction,
        cv=None,
        weight_column=weight,
        output_folder=output_path,
        is_time_series=is_time_series,
    )

    return final_model


def evaluate_data(
    data: pd.DataFrame,
    output_path: Path | str,
    categorical_features: list[str] | None,
    numerical_features: list[str] | None,
    target: str,
    classification_prediction: tuple[int, ...] | None,
    weight: str | None = None,
    allow_transformations: bool = True,
    is_time_series: bool = False,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Data analysis for structured tabular numerical data.

    Where applicable, the following tests are performed:
        - Multicolinearity (VIF)
        - Heteroscedasticity (Breusch-Pagan & White Test)
        - Linearity (Correlation coefficient)
        - Normality (Shapiro-Wilk)
        - Autocorrelation (Durbin-Watson)

    Data transformations (log and scaling) are applied to numerical features
    only if permitted and required (to fix issues the tests reveal).

    Parameters
    ----------
    data
        Pandas Dataframe of your data. Structured or semi-structured
        tabular format. This data should not yet be scaled or encoded.
    output_path
        Path to output location.
    target
        Column in your data that is the target variable (Y, dependent
        variable), what you want to predict.
    weight
        Optional string column value to be used as weight.
    categorical_features
        List of column names (strings) that are categorical variables.
    numerical_features
        List of column names (strings) that are continuous variables.
    classification_prediction
        List of integers that correspond to the target column. The value(s) to
        predict in a classification problem.
    allow_transformations
        Whether to apply transformations.
    is_time_series
        If true then data must be time series. Time series based characteristics
        are taken into consideration during function execution.
    Returns
    -------
    train_transformed
        training data post transformation.
    test_transformed
        test data post transformation.
    """

    if isinstance(output_path, str):
        output_path = Path(output_path)

    if not ValidateData(dataframe=data, custom_index=None, target_column=target).validate():
        raise ValueError(
            "Data not suitable for data analysis. Please run \n"
            "full model flow or tidy_data method prior to \n"
            "data analysis."
        )

    LOG.info(
        "Starting data evaluation for %s prediction",
        "classification" if classification_prediction else "regression",
    )

    # split unscaled data
    train_unscaled, test_unscaled, _ = stratified_split_with_categories(
        df=data,
        categorical_features=categorical_features,
        target_column=target,
        weight_column=weight,
        split_size=None,
        validation_path=None,
        index_columns=None,
        output_path=output_path,
    )

    # encode / scale train
    train_scaled, _, pipeline_out = process_data_pipeline(
        df=train_unscaled,
        numerical_features=numerical_features,
        categorical_features=categorical_features,
        target_column=target,
        weight_column=weight,
        sample_size_encode=True,
        select_encode_values=None,
        encode_values_to_drop=None,
        train_encoded=None,
        test_data=False,
        numerical_pipeline=None,
        output_folder=output_path,
    )

    # encode / scale test
    test_scaled, _, _ = process_data_pipeline(
        df=test_unscaled,
        numerical_features=numerical_features,
        categorical_features=categorical_features,
        target_column=target,
        weight_column=weight,
        sample_size_encode=True,
        select_encode_values=None,
        encode_values_to_drop=None,
        train_encoded=train_scaled,
        test_data=True,
        numerical_pipeline=pipeline_out,
        output_folder=output_path,
    )

    # train/test split for fitting (scaled data)
    x_train, x_test, y_train, y_test, x_train_weight = simple_train_test_split(
        df=train_scaled,
        target_column=target,
        weight_column=weight,
    )

    # initialise model for diagnostics
    if classification_prediction:
        model_initialised = LogisticRegression(max_iter=1000, random_state=42)
    else:
        model_initialised = ElasticNet(random_state=42)

    x_train_model_fit, residuals, _ = initialise_model(
        x_train=x_train,
        x_test=x_test,
        y_train=y_train,
        y_test=y_test,
        x_train_weight=x_train_weight,
        output_folder=output_path,
        model_initialised=model_initialised,
        classification_prediction=classification_prediction,
    )

    # run data analysis
    train_transformed, test_transformed = pre_forecast_data_analysis(
        output_folder=output_path,
        residuals=residuals,
        model_fit=x_train_model_fit,
        model_initialised=model_initialised,
        x_test=x_test,
        y_test=y_test,
        train_scaled=train_scaled,
        test_scaled=test_scaled,
        train_unscaled=train_unscaled,
        test_unscaled=test_unscaled,
        numerical_pipeline=pipeline_out,
        target_column=target,
        weight_column=weight,
        numerical_features=numerical_features,
        categorical_features=categorical_features,
        is_time_series=is_time_series,
        allow_transformations=allow_transformations,
    )

    LOG.info("Data evaluation complete. Results saved to %s", output_path)
    LOG.info("Check 'data_issues_present.csv' for detected issues")

    return train_transformed, test_transformed



# -*- coding: utf-8 -*-
"""
Created on: 1/16/2025
Updated on:

Original author: Ben Taylor
Last update made by:
Other updates made by:

File purpose:

"""

# Built-Ins

# Third Party

# Local Imports
# pylint: disable=import-error,wrong-import-position
# Local imports here
# pylint: enable=import-error,wrong-import-position

# # # CONSTANTS # # #

# # # CLASSES # # #

# # # FUNCTIONS # # #



"""
Created on: 1/16/2025
Original author: Adil Zaheer
"""

# Built-Ins
import logging
import os
from pathlib import Path

# Third Party
import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, mean_squared_error, r2_score
from sklearn.svm import LinearSVC

# Local Imports
from caf.brain.ml._functions.model_selection.functions import (
    calculate_final_coefficients,
)

LOG = logging.getLogger(__name__)


def prediction(
    model,
    test: pd.DataFrame,
    target_column: str | None,
    output_folder: Path,
    validation: pd.DataFrame,
    weight_column: str | None,
    classification_prediction: tuple[int, ...] | None,
    mse: pd.Series,
    drop_vals: pd.DataFrame,
    cols_dropped_by_feat_select: pd.DataFrame,
):
    """
    Generate predictions and final model coefficients, saving results to disk.

    Parameters
    ----------
    model: Fitted final model for prediction on unseen (test) data.
    test: Dataframe of final test data post feature selection.
    target_column: String column name of value to predict.
    output_folder: Path to output location.
    validation: Validation data if available.
    weight_column: Optional string column value to be used as weight.
    classification_prediction: List of integers that correspond to the
                               target column. The value(s) to predict
                               in a classification problem.
    mse: Mean squared error or None. Dependency on if the algorithm selected
         has coefficient values.
    drop_vals: Values dropped during encoding of categorical variables.
    cols_dropped_by_feat_select: These are the columns removed due to feature
                                 selection.

    Returns
    -------
    predictions: Predicted values based on the test data and set to the same
                 index.
    """
    if validation is not None and not target_column:
        raise ValueError("Please provide a target column for prediction as you \
                          have passed a validation set of data. The target column \
                          if a string of the column title.")

    if target_column in test.columns:
        test = test.drop(columns=target_column)

    if weight_column in test.columns:
        weight = test[weight_column].values.flatten()
    else:
        weight = None

    if classification_prediction is not None:
        if validation is not None:
            if isinstance(model, LinearSVC):
                pred_classes = model.predict(test)
                y_true = validation[target_column].values
                accuracy = accuracy_score(y_true, pred_classes, sample_weight=weight)
            else:
                pred_probs = model.predict_proba(test)
                pred_classes = model.classes_[np.argmax(pred_probs, axis=1)]
                y_true = validation[target_column].values
                accuracy = accuracy_score(y_true, pred_classes, sample_weight=weight)

            LOG.info("Accuracy: %s", accuracy)
            accuracy_df = pd.DataFrame({"accuracy": [accuracy]})
            accuracy_df.to_csv(os.path.join(output_folder, "model_performance.csv"))
            predictions = pred_classes
        else:
            if isinstance(model, LinearSVC):
                pred_classes = model.predict(test)
            else:
                pred_probs = model.predict_proba(test)
                pred_classes = model.classes_[np.argmax(pred_probs, axis=1)]
            predictions = pred_classes
    else:
        predictions = model.predict(test)
        if validation is not None:
            r2 = r2_score(validation[target_column], predictions, sample_weight=weight)
            mse = mean_squared_error(
                validation[target_column], predictions, sample_weight=weight
            )
            LOG.info("r2: %s", r2)
            LOG.info("mse: %s", mse)
            metrics_df = pd.DataFrame({"r2": [r2], "mse": [mse]})
            metrics_df.to_csv(os.path.join(output_folder, "model_performance.csv"))

    coeff_df = calculate_final_coefficients(
        model=model,
        test_data=test,
        training_mse=mse,
        predictions=predictions,
        validation_data=validation,
        target_column=target_column,
        is_classification=classification_prediction,
        drop_vals=drop_vals,
        cols_dropped_by_feat_select=cols_dropped_by_feat_select,
    )
    if coeff_df is not None:
        coeff_df.to_csv(
            os.path.join(output_folder, "final_model_coefficients.csv"), index=False
        )

    final_predictions = pd.DataFrame(
        {"predicted_target_column": predictions}, index=test.index
    )
    final_predictions.to_csv(os.path.join(output_folder, "final_predictions.csv"))



"""
Module for model selection _functions. Used to select and initialise a
SciKit-Learn algorithm.
"""

# Built-Ins
import logging
from pathlib import Path
from typing import Optional

# Third Party
import joblib
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import log_loss, mean_squared_error
from sklearn.model_selection import cross_val_score

# Local Imports
from caf.brain.ml._functions._ml_inputs import Models
from caf.brain.ml._functions.process_data_functions.split_data_into_ttv import (
    sample_data,
)

LOG = logging.getLogger(__name__)


def initialise_model(
    x_train: pd.DataFrame,
    x_test: pd.DataFrame,
    y_train: pd.Series,
    y_test: pd.Series,
    output_folder: Path,
    model_initialised: BaseEstimator,
    classification_prediction: tuple[int, ...] | None,
    x_train_weight: pd.DataFrame = None,
) -> tuple[BaseEstimator, pd.Series, float | None]:
    """
    Fit the initialised model and evaluate its initial performance.

    Parameters
    ----------
    x_train: pd.DataFrame of x train values from SciKit-Learns
             train_test_split (simple_train_test_split can be used to
             generate this)
    x_test:  pd.DataFrame of x test values from SciKit-Learns
             train_test_split (simple_train_test_split can be used to
             generate this)
    y_train: pd.Series of y train values from SciKit-Learns
             train_test_split (simple_train_test_split can be used to
             generate this)
    y_test:  pd.Series of y test values from SciKit-Learns
             train_test_split (simple_train_test_split can be used to
             generate this)
    output_folder: Path to output location.
    model_initialised: Initialised SciKitLearn model.
    classification_prediction: List of integers that correspond to the
                               target column. The value(s) to predict
                               in a classification problem.
    x_train_weight:  Numpy ndarray of weight values that correspond to
                     x_train generated in simple_train_test_split

    Returns
    -------
    model_fit: Fitted model on train_test_split test data.
    residuals: Truth values form the train_test_split against the predictions.
    mse: Mean squared error of predictions.
    """
    weight = None
    if x_train_weight is not None:
        weight = x_train_weight.values.flatten()

    model_filename = output_folder / "initial_fitted_model.pkl"
    if model_filename.exists():
        model_fit = joblib.load(model_filename)
    else:
        if weight is not None:
            model_fit = model_initialised.fit(x_train, y_train, sample_weight=weight)
        else:
            model_fit = model_initialised.fit(x_train, y_train)

        joblib.dump(model_fit, model_filename)

    y_pred = model_fit.predict(x_test)
    residuals = y_test - y_pred
    coeff_df, mse = calculate_model_coeff(
        model=model_fit,
        x_train=x_train,
        x_test=x_test,
        y_test=y_test,
        residuals=residuals,
        classification_prediction=classification_prediction,
        y_pred=y_pred,
    )

    if coeff_df is not None:
        coeff_df.to_csv(output_folder / "initial_model_coefficients.csv", index=False)

    return model_fit, residuals, mse


def select_model(
    train: pd.DataFrame,
    output_folder: Path,
    target_column: str | None,
    weight_column: str | None,
    models_to_test: list[Models],
    classification_prediction: tuple[int, ...] | None,
    is_time_series: bool = False,
) -> BaseEstimator:
    """
    Quickly assess and select the best model from a list of candidates.

    Parameters
    ----------
    train: Processed input data split into train subset.
    output_folder: Path to output location.
    target_column: Sting column name of value to predict.
    weight_column: Optional string column value to be used as weight.
    models_to_test: List or one algorithm to use as the base of the model.
                    Available algorithms can be seen in
                    _ml_inputs.py or __info__.py.
    classification_prediction: List of integers that correspond to the
                               target column. The value(s) to predict
                               in a classification problem.
    is_time_series: If true data is temporal is nature.

    Returns
    -------
    best_model: Best performing model initialised.
    """
    eval_df_path = Path(output_folder) / "model_algorithm_evaluation.csv"

    if eval_df_path.exists():
        LOG.info(
            "model_algorithm_evaluation.csv already exists so the best "
            "model is being selected from these results."
        )
        df = pd.read_csv(eval_df_path)
        if classification_prediction:
            best_idx = df["F1"].idxmax()
            best_model_str = str(df.loc[best_idx, "Models"])
        else:
            best_idx = df["R2"].idxmax()
            best_model_str = str(df.loc[best_idx, "Models"])
        enum_name = best_model_str.split(".")[1]
        best_model_enum = Models[enum_name]
        best_model = best_model_enum.get_model()

        return best_model

    if not target_column:
        raise ValueError(
            "Please provide a target column. "
            "This should be a column title passed as a string."
        )

    x = train.drop(columns=[target_column] + ([weight_column] if weight_column else []))
    y = train[target_column]
    weight = train[weight_column].values.flatten() if weight_column else None

    acc = {}
    best_score = float("-inf")
    best_model = None

    x, y, weight = sample_data(x=x, y=y, weight=weight, is_time_series=is_time_series)

    for model_enum in models_to_test:
        # scikit
        model_instance = model_enum.get_model()
        LOG.info("Testing model: %s", model_instance)

        if isinstance(model_instance, LogisticRegression):
            model_instance.set_params(max_iter=1000)

        if classification_prediction:
            scores_f1, scores_auc = score_classification(
                weight=weight, model_instance=model_instance, x=x, y=y
            )
            mean_score = scores_f1.mean()
            acc[model_enum] = {"F1": scores_f1.mean(), "AUC": -scores_auc.mean()}
        else:
            scores_r2, scores_mse = score_regression(
                weight=weight, model_instance=model_instance, x=x, y=y
            )
            mean_score = scores_r2.mean()
            acc[model_enum] = {"R-squared": scores_r2.mean(), "MSE": scores_mse.mean()}

        if mean_score > best_score:
            best_score = mean_score
            best_model = model_enum.get_model()

    LOG.info("Best model: %s", best_model)
    LOG.info("Best model score: %s", best_score)
    evaluation_df = pd.DataFrame.from_dict(acc, orient="index")
    evaluation_df.index.name = "Models"
    evaluation_df.to_csv(output_folder / "model_algorithm_evaluation.csv", index=True)

    return best_model


def score_regression(
    weight: pd.DataFrame, model_instance: Models, x: pd.DataFrame, y: pd.DataFrame
):
    """
    Score regression models using cross-validation.

    Parameters
    ----------
    weight: Pandas dataframe of weight values from the original
            train input data.
    model_instance: Initialised model.
    x: Train data split into only the explanatory variables. Target
       and weight should be removed. Any index columns should be set.
    y: Train data split into only the target. Any index columns should be set.

    Returns
    -------
    scores_r2: Series of R2 scores.
    scores_mse: Series of mean squared error scores.
    """
    if weight is not None:
        scores_r2 = cross_val_score(
            model_instance,
            x,
            y,
            cv=3,
            scoring="r2",
            n_jobs=-1,
            params={"sample_weight": weight},
            verbose=1,
        )
        scores_mse = -cross_val_score(
            model_instance,
            x,
            y,
            cv=3,
            scoring="neg_mean_squared_error",
            n_jobs=-1,
            params={"sample_weight": weight},
            verbose=1,
        )
    else:
        scores_r2 = cross_val_score(
            model_instance, x, y, cv=3, scoring="r2", n_jobs=-1, verbose=1
        )
        scores_mse = -cross_val_score(
            model_instance, x, y, cv=3, scoring="neg_mean_squared_error", n_jobs=-1, verbose=1
        )

    return scores_r2, scores_mse


def score_classification(
    weight: pd.DataFrame, model_instance: Models, x: pd.DataFrame, y: pd.DataFrame
):
    """
    Score classification models using cross-validation.

    Parameters
    ----------
    weight: Pandas dataframe of weight values from the original train input data.
    model_instance: Initialised model.
    x: Train data split into only the explanatory variables. Target
       and weight should be removed. Any index columns should be set.
    y: Train data split into only the target. Any index columns should be set.

    Returns
    -------
    scores_f1: Series of F1 scores.
    scores_auc: Series of AUC scores.
    """
    y_ = y.squeeze()
    if y_.nunique() > 2:
        f1 = "f1_weighted"
        roc_auc = "roc_auc_ovr_weighted"
    else:
        f1 = "f1"
        roc_auc = "roc_auc"

    if weight is not None:
        scores_f1 = cross_val_score(
            model_instance,
            x,
            y,
            cv=3,
            scoring=f1,
            n_jobs=-1,
            params={"sample_weight": weight},
            verbose=1,
        )
        scores_auc = -cross_val_score(
            model_instance,
            x,
            y,
            cv=3,
            scoring=roc_auc,
            n_jobs=-1,
            params={"sample_weight": weight},
            verbose=1,
        )
    else:
        scores_f1 = cross_val_score(
            model_instance, x, y, cv=3, scoring=f1, n_jobs=-1, verbose=1
        )
        scores_auc = -cross_val_score(
            model_instance, x, y, cv=3, scoring=roc_auc, n_jobs=-1, verbose=1
        )

    return scores_f1, scores_auc


def calculate_model_coeff(
    model,
    x_train: pd.DataFrame,
    x_test: pd.DataFrame | None,
    y_test: pd.Series | None,
    residuals: pd.Series | None,
    classification_prediction: tuple[int, ...] | None,
    y_pred: pd.Series,
) -> tuple[Optional[pd.DataFrame], Optional[float]]:
    """
    Calculate model coefficients and statistics.

    Parameters
    ----------
    model:
        Fitted model on train_test_split of training data.
    x_train:
        Series of train data to be used as train.
    x_test:
        Series of test data to be used as unseen test data.
    y_test:
        Series of target column inside train to be used as validation
        for predictions.
    residuals:
        Series of residual values based on x_test predictions.
    classification_prediction:
        List of integers that correspond to the target column. The value(s) to
        predict in a classification problem.
    y_pred:
        Series of predicted values based on training data.

    Returns
    -------
    coeff_df:
        Dataframe of coefficient values and other relevant statistics.
    mse:
        Mean squared error of predictions.
    """
    if _is_statsmodels_model(model):
        return _extract_statsmodels_inference(model, x_test, y_test)

    if hasattr(model, "coef_"):
        if len(model.coef_.shape) == 2 and model.coef_.shape[0] > 1:
            return None, None

        return _extract_sklearn_coefficients(
            model, x_train, x_test, y_test, y_pred, residuals, classification_prediction
        )

    if hasattr(model, "feature_importances_"):
        return _extract_feat_importance(
            model, x_train, x_test, y_test, y_pred, classification_prediction
        )

    LOG.warning(
        "Model %s does not support coefficient or importance extraction",
        {type(model).__name__},
    )
    return None, None


def calculate_final_coefficients(
    model,
    test_data: pd.DataFrame,
    training_mse: pd.Series,
    predictions: pd.Series,
    validation_data: pd.DataFrame,
    target_column: str | None,
    is_classification: tuple[int, ...] | None,
    drop_vals: pd.DataFrame,
    cols_dropped_by_feat_select: pd.DataFrame,
) -> pd.DataFrame | None:
    """
    Calculate final model coefficients and statistics.

    Parameters
    ----------
    model:
        Fitted final model for prediction on unseen (test) data.
    test_data:
        Dataframe of final test data post feature selection.
    training_mse:
        Mean squared error of predictions based on training data.
    predictions:
        Predicted values based on the test data and set to the same index.
    validation_data:
        Validation data if available. Must pass target column if passing
        validation data otherwise validation will not be used.
    target_column:
        String column name of value to predict.
    is_classification:
        List of integers that correspond to the target column. The value(s) to
        predict in a classification problem.
    drop_vals:
        Values dropped during encoding of categorical variables.
    cols_dropped_by_feat_select:
        These are the columns removed due to feature selection.

    Returns
    -------
    coeff_df:
        Dataframe of coefficient values and other relevant statistics.
    """
    if validation_data is not None and target_column is not None:
        coeff_df, error_metric = calculate_model_coeff(
            model=model,
            x_train=test_data,  # only using it for the column names
            x_test=test_data,
            y_test=validation_data[target_column],
            y_pred=predictions,
            residuals=None,
            classification_prediction=is_classification,
        )
    else:
        coeff_df, error_metric = calculate_model_coeff(
            model=model,
            x_train=test_data,
            x_test=test_data,
            y_test=None,
            y_pred=predictions,
            residuals=None,
            classification_prediction=is_classification,
        )
        error_metric = training_mse if error_metric is None else error_metric

    if coeff_df is None:
        return None

    coeff_df["Error_Metric"] = error_metric
    coeff_df["Error_Source"] = "validation" if validation_data is not None else "training"

    base_columns = list(coeff_df.columns)
    dropped_dfs = []

    if drop_vals is not None and len(drop_vals.columns) > 0:
        drop_vals_features = pd.DataFrame(
            {
                "Feature": drop_vals.columns,
                "Status": "Dropped during encoding",
            }
        )
        for col in base_columns:
            if col not in drop_vals_features.columns:
                drop_vals_features[col] = "N/A"
        dropped_dfs.append(drop_vals_features)

    if (
        cols_dropped_by_feat_select is not None
        and len(cols_dropped_by_feat_select.columns) > 0
    ):
        feat_select_features = pd.DataFrame(
            {
                "Feature": cols_dropped_by_feat_select.columns,
                "Status": "Dropped during feature selection",
            }
        )
        for col in base_columns:
            if col not in feat_select_features.columns:
                feat_select_features[col] = "N/A"
        dropped_dfs.append(feat_select_features)

    if dropped_dfs:
        coeff_df["Status"] = "Active in model"
        coeff_df = pd.concat([coeff_df] + dropped_dfs, ignore_index=True)

    return coeff_df


def _extract_sklearn_coefficients(
    model,
    x_train: pd.DataFrame,
    x_test: pd.DataFrame | None,
    y_test: pd.Series | None,
    y_pred: pd.Series | None,
    residuals: Optional[pd.Series] | None,
    classification_prediction: Optional[tuple[int, ...]] = None,
) -> tuple[Optional[pd.DataFrame], Optional[float]]:
    """
    Extract coefficients from scikit-learn linear models.

    No p-values or standard errors (not mathematically valid for sklearn).

    Valid for:
    - LinearRegression
    - Ridge, Lasso, ElasticNet
    - LogisticRegression (L1, L2, ElasticNet)
    - LinearSVC

    Parameters
    ----------
    model:
        Fitted model on train_test_split of training data.
    x_train:
        Series of train data to be used as train.
    x_test:
        Series of test data to be used as unseen test data.
    y_test:
        Series of target column inside train to be used as validation
        for predictions.
    residuals:
        Series of residual values based on x_test predictions.
    classification_prediction:
        List of integers that correspond to the target column. The value(s) to
        predict in a classification problem.
    y_pred:
        Series of predicted values based on training data.

    Returns
    -------
    coeff_df:
        Dataframe of coefficient values and other relevant statistics.
    error_metric:
        Mean squared error of predictions.
    """
    # multinomial
    if len(model.coef_.shape) == 2 and model.coef_.shape[0] > 1:
        LOG.info("Multinomial logistic regression excluded")
        return None, None

    if hasattr(model, "intercept_"):
        intercept = np.array(model.intercept_).flatten()
        coefficients = np.array(model.coef_).flatten()
        coefficients = np.concatenate([intercept, coefficients])
        feature_names = ["intercept"] + list(x_train.columns)
    else:
        coefficients = model.coef_.flatten()
        feature_names = list(x_train.columns)

    coeff_df = pd.DataFrame(
        {
            "Feature": feature_names,
            "Coefficient": coefficients,
            "Abs_Coefficient": np.abs(coefficients),
        }
    )

    if isinstance(model, LogisticRegression):
        coeff_df["Odds_Ratio"] = np.exp(coefficients)
        coeff_df["Note"] = (
            "Odds ratios from sklearn LogisticRegression (no p-values available)"
        )
    else:
        coeff_df["Note"] = "Coefficients from sklearn (no statistical inference available)"

    coeff_df = coeff_df.sort_values("Abs_Coefficient", ascending=False)

    if y_test is not None:
        if classification_prediction:
            if hasattr(model, "predict_proba"):
                proba = model.predict_proba(x_test)
                if proba.shape[1] == 2:
                    # binary
                    error_metric = log_loss(y_test, proba[:, 1])
                else:
                    # multiclass
                    error_metric = log_loss(y_test, proba)
            else:
                # linear svc
                error_metric = log_loss(y_test, y_pred, labels=np.unique(y_test))
        else:
            # regression
            error_metric = (
                np.mean(residuals**2)
                if residuals is not None
                else mean_squared_error(y_test, y_pred)
            )
        LOG.info("Extracted sklearn coefficients for %s features", len(coeff_df))
        return coeff_df, error_metric

    LOG.info("Extracted sklearn coefficients for %s features", len(coeff_df))
    return coeff_df, None


def _extract_feat_importance(
    model,
    x_train: pd.DataFrame,
    x_test: pd.DataFrame | None,
    y_test: pd.Series | None,
    y_pred: pd.Series | None,
    classification_prediction: Optional[tuple[int, ...]] = None,
) -> tuple[pd.DataFrame, Optional[float]]:
    """
    Extract feature importances from tree-based models (sklearn).

    Valid for:
    - RandomForest (Classifier/Regressor)
    - ExtraTrees (Classifier/Regressor)
    - GradientBoosting (Classifier/Regressor)
    - DecisionTree (Classifier/Regressor)
    - AdaBoost
    - Bagging

    Parameters
    ----------
    model:
        Fitted model on train_test_split of training data.
    x_train:
        Series of train data to be used as train.
    x_test:
        Series of test data to be used as unseen test data.
    y_test:
        Series of target column inside train to be used as validation
        for predictions.
    classification_prediction:
        List of integers that correspond to the target column. The value(s) to
        predict in a classification problem.
    y_pred:
        Series of predicted values based on training data.

    Returns
    -------
    importance_df:
        Dataframe of importance values.
    error_metric:
        Log loss or mean squared error of predictions.
    """
    importance_df = pd.DataFrame(
        {
            "Feature": x_train.columns,
            "Importance": model.feature_importances_,
            "Importance_Type": "Gini/MDI",
        }
    ).sort_values("Importance", ascending=False)

    importance_df["Cumulative_Importance"] = importance_df["Importance"].cumsum()

    if y_test is not None:
        if classification_prediction:
            if hasattr(model, "predict_proba"):
                proba = model.predict_proba(x_test)
                error_metric = log_loss(y_test, proba)
            else:
                error_metric = log_loss(y_test, y_pred)
        else:
            error_metric = mean_squared_error(y_test, y_pred)

        LOG.info("Extracted feature importances for %s features", len(importance_df))
        return importance_df, error_metric
    LOG.info("Extracted feature importances for %s features", len(importance_df))
    return importance_df, None


def _is_statsmodels_model(model) -> bool:
    """
    Check if model is statsmodels model.

    Parameters
    ----------
    model:
        Fitted model on train_test_split of training data.
    Returns
    -------
    True if statsmodels algorithm.
    """
    return any(base.__module__.startswith("statsmodels") for base in model.__class__.__mro__)


def _extract_statsmodels_inference(
    model,
    x_test: pd.DataFrame | None,
    y_test: pd.Series | None,
    classification_prediction: Optional[tuple[int, ...]] = None,
) -> tuple[Optional[pd.DataFrame], Optional[float]]:
    """
    Extract full statistical inference from statsmodels.

    Valid for:
    - statsmodels.OLS (returns t-values)
    - statsmodels.Logit (returns z-values)
    - statsmodels.GLM (returns z-values)

    Parameters
    ----------
    model:
        Fitted model on train_test_split of training data.
    x_test:
        Series of test data to be used as unseen test data.
    y_test:
        Series of target column inside train to be used as validation
        for predictions.
    classification_prediction:
        List of integers that correspond to the target column. The value(s) to
        predict in a classification problem.

    Returns
    -------
    coeff_df:
        Dataframe of coefficient values and other relevant statistics.
    error_metric:
        Mean squared error or log loss of predictions.
    """
    # multinomial
    if "MNLogit" in str(model.__class__):
        LOG.info("Multinomial logit results excluded")
        return None, None

    try:
        stats_df = pd.DataFrame(
            {
                "Feature": model.params.index,
                "Coefficient": model.params.values,
                "Std_Error": model.bse.values,
                "Statistic": model.tvalues.values,  # t-value for OLS, z-value for Logit/GLM
                "P_Value": model.pvalues.values,
                "CI_Lower_95": model.conf_int()[0].values,
                "CI_Upper_95": model.conf_int()[1].values,
            }
        )

        if "Logit" in str(model.__class__):
            stats_df["Odds_Ratio"] = np.exp(stats_df["Coefficient"])
            stats_df["OR_CI_Lower_95"] = np.exp(stats_df["CI_Lower_95"])
            stats_df["OR_CI_Upper_95"] = np.exp(stats_df["CI_Upper_95"])

        if y_test is not None:
            if hasattr(model, "predict"):
                if classification_prediction:
                    # logistic regression
                    y_pred_proba = model.predict(x_test)
                    error_metric = log_loss(y_test, y_pred_proba)
                else:
                    # OLS
                    y_pred = model.predict(x_test)
                    error_metric = mean_squared_error(y_test, y_pred)
                LOG.info("Extracted statsmodels inference with %s features", len(stats_df))
                return stats_df, error_metric
        else:
            LOG.info("Extracted statsmodels inference with %s features", len(stats_df))
        return stats_df, None

    except Exception as e:
        LOG.error("Failed to extract statsmodels inference: %s", e)
        return None, None
